import aton.spx as spx
from pytest import approx


folder = 'tests/samples/'
example_csv = folder + 'example_spx_2.02g.csv'
example_csv_nd = folder + 'example_spx_ND_1.284g.csv'


def test_spectra():
    # Define the Spectra data and Plotting options
    ins = spx.Spectra(
        type     = 'INS',
        files    = [example_csv, example_csv_nd],
        units_in = 'cm1',
        units    = 'meV',
        plotting = spx.Plotting(
            title      = 'Example spectra',
            xlim       = [8, 1000],
            offset     = True,
            scaling    = 0.9,
            margins    = [0.2, 0.2],
            xlabel     = 'Energy / meV',
            ylabel     = 'S(Q,E)',
            legend     = ['example 1', 'example 2'],
            log_xscale = True,
        ),
    )
    assert ins.units == ['meV', 'meV']
    assert ins.type == 'ins'


def test_normalize():
    # Define the Spectra data and Plotting options
    ins = spx.Spectra(
        type     = 'ins',
        files    = [example_csv, example_csv_nd],
        units_in = 'cm-1',
        units    = 'meV',
        plotting = spx.Plotting(
            title      = 'Example spectra',
            xlim       = [8, 1000],
            offset     = True,
            scaling    = 0.9,
            margins    = [0.2, 0.2],
            xlabel     = 'Energy / meV',
            ylabel     = 'S(Q,E)',
            legend     = ['example 1', 'example 2'],
            log_xscale = True,
        ),
    )
    # Normalise the spectra to the same height in a given range
    normalization_range = [5, 500]
    ins = spx.normalize.height(spectra=ins, range=normalization_range)
    # The full range was not normalised
    assert max(ins.dfs[0]['S(Q,E)']) != max(ins.dfs[1]['S(Q,E)'])
    # The normalised range should have the same max
    trimmed_dfs = []
    for df in ins.dfs:
        df = df[(df[df.columns[0]] >= normalization_range[0])]
        df = df[(df[df.columns[0]] <= normalization_range[1])]
        trimmed_dfs.append(df)
    ins.dfs = trimmed_dfs
    assert max(ins.dfs[0]['S(Q,E)']) == max(ins.dfs[1]['S(Q,E)'])


def test_fit():
    ins = spx.Spectra(
        type     = 'ins',
        files    = [example_csv, example_csv_nd],
        units_in = 'cm-1',
        units    = 'meV',
    )
    plateau, plateau_error = spx.fit.plateau(spectra=ins, cuts=[33, 36])
    assert float(plateau) == approx(0.14, abs=0.01)
    assert float(plateau_error) == approx(0.01, abs=0.01)
    area, area_error = spx.fit.area_under_peak(spectra=ins, peak=[36, 39, plateau, plateau_error])
    assert float(area) == approx(1.12, abs=0.02)
    assert float(area_error) == approx(0.02, abs=0.01)




"""
# Description

Functions to search for specific content inside text files.


# Index

Find and return specific text strings from a file  
`lines()`  
`between()`  

Find the `(start, end)` position of specific strings in a file  
`pos()`  
`pos_regex()`  
`next_pos()`  
`next_pos_regex()`  
`line_pos()`  
`between_pos()`  


# Examples

To find the lines containing the word 'key', plus an additional line below,
```python
from aton import text
# with split = False
text.find.lines(filepath=file, key='key', additional=1)
    # ['line with key 1\\nline below first match',
    #  'line with key 2\\nline below second match]
# with split = True
text.find.lines(filepath=file, key='key', additional=1, split=True)
    # ['line with key 1',
    # 'line below first match',
    # 'line with key 2',
    # 'line below second match]
```

To find the text between the lines containing the words 'first' and 'second',
```python
from aton import text
text.find.between(filepath=file, key1='first', key2='second')
    # 'first line\\nadditional\\nlines\\nin\\nbetween\\nsecond line'
```

---
"""


import mmap
import re
import aton.st.file as file


def lines(
        filepath:str,
        key:str,
        matches:int=0,
        additional:int=0,
        split: bool=False,
        regex:bool=False
    ) -> list:
    """Returns a list with the matches containing the `key` string in `filepath`.

    If no match is found, returns an empty list.

    To use regular expressions in the search, set `regex=True`
    (deactivated by default).

    The value `matches` specifies the max number of matches to be returned.
    Defaults to 0 to return all possible matches. Set it to 1 to return only one match,
    or to negative integers to start the search from the end of the file upwards.

    The value `additional` specifies the number of additional lines
    below the target line that are also returned;
    2 to return the found line plus two additional lines below, etc.
    Negative values return the specified number of lines before the target line.
    The original ordering from the file is preserved.
    Defaults to `additional=0`, only returning the target line.
    By default, the additional lines are returned
    in the same list item as the match separated by a `\\n`,
    unless `split=True`, in which case these additional lines
    are splitted and added as additional items in the list.
    """
    file_path = file.get(filepath)
    matches_found = []
    if regex:
        positions = pos_regex(file_path, key, matches)
    else:
        positions = pos(file_path, key, matches)
    with open(file_path, 'r+b') as f:
        mm = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
    for start, end in positions:
        # Get the positions of the full line containing the match
        line_start = mm.rfind(b'\n', 0, start) + 1
        line_end = mm.find(b'\n', end, len(mm))
        # Default values for the start and end of the line
        if line_start == -1: line_start = 0
        if line_end == -1: line_end = len(mm)
        # Adjust the line_end to add additional lines after the match
        match_start = line_start
        match_end = line_end
        if additional > 0:
            for _ in range(abs(additional)):
                match_end = mm.find(b'\n', match_end + 1, len(mm)-1)
                if match_end == -1:
                    match_end = len(mm)
                    break
        elif additional < 0:
            for _ in range(abs(additional)):
                match_start = mm.rfind(b'\n', 0, match_start - 1) + 1
                if match_start == -1:
                    match_start = 0
                    break
        # Save the matched lines
        matches_found.append(mm[match_start:match_end].decode())
    if split:
        splitted_matches_found = []
        for string in matches_found:
            splitted_match = string.splitlines()
            splitted_matches_found.extend(splitted_match)
        matches_found = splitted_matches_found
    return matches_found


def between(
        filepath:str,
        key1:str,
        key2:str,
        include_keys:bool=True,
        match:int=1,
        regex:bool=False
    ) -> str:
    """Returns the content between the lines with `key1` and `key2` in `filepath`.

    Keywords can be at any position within the line.
    Regular expressions can be used by setting `regex=True`.

    Key lines are omited by default, but can be returned with `include_keys=True`.

    If there is more than one match, only the first one is considered by default;
    set `match` (int) to specify a particular match (1, 2... 0 is considered as 1!).
    Use negative numbers to start from the end of the file.

    If no match is found, returns an empty string.

    If `key2` is not found, it returns all the text from `key1` to the end of the file.
    """
    file_path = file.get(filepath)
    start, end = between_pos(file_path, key1, key2, include_keys, match, regex)
    if (start, end) == (-1, -1):
        return ''
    with open(file_path, 'r+b') as f:
        mm = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
    return (mm[start:end].decode())


def pos(
        filepath,
        key:str,
        matches:int=0
        ) -> list:
    """Returns a list with the positions of the `key` in `filepath`.

    If no match is found, returns an empty list.

    The `filepath` can be a file or a memory mapped file.

    The value `matches` specifies the max number of matches to return.
    Defaults to 0 to return all possible matches.
    Set it to 1 to return only one match,
    2 to get the first two matches, etc.
    You can also set it to negative integers to start
    searching from the end of the file upwards.

    This method is faster than `pos_regex()`,
    but does not search for regular expressions.
    """
    positions = []
    mm = filepath
    if not isinstance(filepath, mmap.mmap):
        file_path = file.get(filepath)
        with open(file_path, 'r+b') as f:
            mm = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
    keyword_bytes = key.encode()
    if matches >= 0:
        start = 0
        while matches == 0 or len(positions) < matches:
            pos = mm.find(keyword_bytes, start)
            if pos == -1:
                break
            end = pos + len(keyword_bytes)
            positions.append((pos, end))
            start = end
    else:
        start = len(mm)
        while len(positions) < abs(matches):
            pos = mm.rfind(keyword_bytes, 0, start)
            if pos == -1:
                break
            end = pos + len(keyword_bytes)
            positions.append((pos, end))
            start = pos
        positions.reverse()
    return positions


def pos_regex(
        filepath,
        key:str,
        matches:int=0
    ) -> list:
    """Returns a list of the positions of a `key` in a given `filepath` (actual file, not mmapped!).

    The value `matches` specifies the max number of matches to return.
    Defaults to 0 to return all possible matches. Set it to 1 to return only one match,
    or to negative integers to start searching from the end of the file upwards.

    For big files, this method is slower than `pos()`, but it can search for regular expressions.
    """
    file_path = file.get(filepath)
    positions = []
    with open(file_path, 'r', encoding='utf-8') as f:
        content = f.read()
    if matches > 0:
        start = 0
        while len(positions) < matches:
            match = re.search(key, content[start:])
            if not match:
                break
            match_start = start + match.start()
            match_end = start + match.end()
            positions.append((match_start, match_end))
            start = match_end
    else:
        all_matches = list(re.finditer(key, content))
        if matches == 0:
            positions = [(match.start(), match.end()) for match in all_matches]
        else:
            positions = [(match.start(), match.end()) for match in all_matches[-abs(matches):]]
    return positions


def next_pos(
        filepath,
        position:tuple,
        key:str,
        match:int=1
    ) -> tuple:
    """Get the next position of the `key` in the `filepath` (file or mmapped file), starting from an initial `position` tuple.

    The `match` number specifies the nonzero index of the next match to return (1, 2... 0 is considered as 1!).
    It can be negative to search backwards from the initial position.
    The last known positions will be returned if no more matches are found.

    This method is specific for normal strings.
    To use regular expressions, check `next_pos_regex()`.
    """
    mm = filepath
    if not isinstance(filepath, mmap.mmap):
        file_path = file.get(filepath)
        with open(file_path, 'r+b') as f:
            mm = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
    start, end = position
    keyword_bytes = key.encode()
    if match == 0:
        match = 1
    positions = []
    if match > 0:
        for _ in range(match):
            start = mm.find(keyword_bytes, end, len(mm))
            if start == -1:
                break
            end = start + len(keyword_bytes)
            positions.append((start, end))
    else:
        for _ in range(abs(match)):
            start = mm.rfind(keyword_bytes, 0, start)
            if start == -1:
                break
            end = start + len(keyword_bytes)
            positions.append((start, end))
    positions.reverse()
    if len(positions) == 0:               ###########  TODO  what should this return?? 
        positions.append((-1, -1))
    return positions[0]                     #####  should return empty if not found?


def next_pos_regex(
        filepath,
        position:tuple,
        key:str,
        match:int=0
    ) -> tuple:
    """Get the next position of the `key` in the `filepath` (actual file, not mmapped!), starting from an initial `position` tuple.

    The `match` number specifies the next match to return (1, 2... 0 is considered as 1!).
    It can be negative to search backwards from the initial position.
    This method is specific for regular expressions.

    For normal strings, check the faster `next_pos()` method.
    """
    file_path = file.get(filepath)
    start, end = position
    with open(file_path, 'r') as f:
        content = f.read()
    if match == 0:
        match = 1
    positions = []
    if match > 0:
        for _ in range(match):
            match_found = re.search(key, content[end:])
            if not match_found:
                break
            start = end + match_found.start()
            end = end + match_found.end()
            positions.append((start, end))
        positions.reverse()
        if len(positions) == 0:  ###########  when pos did not find, it was []  !! should be same here?
            positions.append((-1, -1))
        return positions[0]
    else:  # Reverse match
        all_matches = list(re.finditer(key, content))
        if not all_matches:  ###########   when pos did not find, it was []  !! should be same here?
            return (-1, -1)
        if abs(match) > len(all_matches):
            match = -len(all_matches)
        else:
            match_found = all_matches[match]  # Already negative
            start = match_found.start()
            end = match_found.end()
    return start, end


def line_pos(
        filepath,
        position:tuple,
        skips:int=0
    ) -> tuple:
    """Get the position of the full line containing the `position` tuple in `filepath` (whether file or memory mapped file).

    A specific line below can be returned with `skips` being a natural int,
    or previous lines with negative values.
    """
    mm = filepath
    if not isinstance(filepath, mmap.mmap):
        file_path = file.get(filepath)
        with open(file_path, 'r+b') as f:
            mm = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
    if position == (-1, -1):  # No match
        return (-1, -1)
    start, end = position
    if skips == 0:
        start = mm.rfind(b'\n', 0, start) + 1
        end = mm.find(b'\n', end, len(mm))
    elif skips > 0:
        for i in range(0, abs(skips)):
            start = mm.find(b'\n', end, len(mm)) + 1
            if start == -1:
                start = len(mm)
                end = len(mm)
                break
            end = mm.find(b'\n', start, len(mm))
            if end == -1:
                start = len(mm)
                end = len(mm)
                break
    else:  # previous lines
        for i in range(0, abs(skips)):
            end = mm.rfind(b'\n', 0, start)
            if end == -1:
                start = 0
                end = 0
                break
            start = mm.rfind(b'\n', 0, end) + 1
            if start == -1:
                start = 0
                end = 0
                break
    return start, end


def between_pos(
        filepath,
        key1:str,
        key2:str,
        include_keys:bool=True,
        match:int=1,
        regex:bool=False
    ) -> tuple:
    """Returns the positions of the content between the lines containing `key1` and `key2` in the `filepath`.

    Keywords can be at any position within the line.
    Regular expressions can be used by setting `regex=True`.

    Key lines are omited by default, but can be returned with `include_keys=True`.

    If there is more than one match, only the first one is considered by default;
    set `match` number to specify a particular match (1, 2... 0 is considered as 1!).
    Use negative numbers to start from the end of the file.

    If `key2` is not found, it returns the text position from `key1` to the end of the file.
    """
    file_path = file.get(filepath)
    if match == 0:
        match = 1
    if regex:
        positions_1: list = pos_regex(file_path, key1, match)
        if not positions_1:
            return (-1, -1)
        if match > 0:
            positions_1.reverse()
        position_1 = positions_1[0]
        position_2: tuple = next_pos_regex(file_path, position_1, key2, 1)
    else:
        positions_1: list = pos(file_path, key1, match)
        if not positions_1:
            return (-1, -1)
        if match > 0:
            positions_1.reverse()
        position_1 = positions_1[0]
        position_2: tuple = next_pos(file_path, position_1, key2, 1)
    skip_line_1 = 0
    skip_line_2 = 0
    if not include_keys:
        skip_line_1 = 1
        skip_line_2 = -1
    with open(file_path, 'r+b') as f:
        mm = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)
    start, _ = line_pos(mm, position_1, skip_line_1)
    if position_2 != (-1, -1):
        _, end = line_pos(mm, position_2, skip_line_2)
    else:
        end = len(mm)
    return start, end




from aton.st import file


folder = 'tests/samples/'
sample = folder + 'sample.txt'
sample_copy = folder + 'sample_copy.txt'
sample_copy_2 = folder + 'sample_copy_2.txt'
sample_ok = folder + 'sample_ok.txt'
sample_ok_2 = folder + 'sample_ok_2.txt'


def test_get():
    # Clean from previous tests
    try:
        file.remove(sample_copy)
    except:
        pass
    # Finds an existing file
    try:
        assert file.get(sample) != None
        assert True
    except FileNotFoundError:
        assert False
    # Does not find a non-existing file
    try:
        file.get(sample_copy)
        assert False
    except FileNotFoundError:
        assert True
    # get_list, 'tests/sample.txt' in 'fullpath/tests/sample.txt'
    try:
        assert sample in file.get_list(folder, filters='sample')[0]
    except:
        assert False


def test_copy():
    try:
        file.remove(sample_copy)
        file.remove(sample_copy_2)
    except:
        pass
    # Copy files
    try:
        file.copy(sample, sample_copy)
        assert file.get(sample_copy) != None
    except FileNotFoundError:
        assert False
    # Move files
    try:
        file.move(sample_copy, sample_copy_2)
        assert file.get(sample_copy_2) != None
    except:
        assert False
    try:
        file.get(sample_copy)
        assert False
    except FileNotFoundError:
        assert True
    # Remove
    try:
        file.remove(sample_copy_2)
    except:
        assert False
    try:
        file.get(sample_copy_2)
        assert False
    except:
        assert True


def test_rename():
    try:
        file.remove(sample_copy)
        file.remove(sample_ok)
    except:
        pass
    file.copy(sample, sample_copy)
    file.rename_on_folder(old='copy', new='ok', folder=folder)
    try:
        file.remove(sample_ok)
        assert True
    except:
        assert False
    try:
        file.remove(sample_copy)
        assert False
    except:
        assert True


def test_folders():
    try:
        file.remove(sample_copy)
        file.remove(sample_copy_2)
        file.remove(sample_ok)
        file.remove(sample_ok_2)
    except:
        pass
    file.copy_to_folders(extension='.txt', strings_to_delete=['.txt'], folder=folder)
    try:
        assert file.get_list(folder=folder+'sample', abspath=False) == ['sample.txt']
    except:
        assert False
    # Check that the folder is deleted
    file.remove(folder+'sample')
    try:
        file.get_list(folder+'sample')
        assert False
    except FileNotFoundError:
        assert True




"""
# Description

Functions to handle bash calls and related operations on Linux systems.


# Index

| | |  
| --- | --- |  
| `bash()` | Run a bash shell commands |  
| `git()` | Automatically update a Git repository |  
| `here()` | Runs the rest of the script inside a given folder |  

---
"""


import subprocess
import datetime
import sys
import os
from aton._version import __version__


def bash(
        command:str,
        cwd=None,
        verbose:bool=True,
        return_anyway:bool=False
    ):
    """Run a bash shell commands.
    
    A given `command` will be executed inside an optional `cwd` directory.
    If empty, the current working directory will be used.
    Prints the running command and outputs by default, override this with `verbose=False`.
    Returns the result of the command used, except for when
    errors are raised automatically; set `return_anyway=True` to override this.
    """
    if verbose:
        print(f'$ {command}')
    result = subprocess.run(command, cwd=cwd, shell=True, text=True, capture_output=True)
    if verbose and result.returncode == 0 and result.stdout:
        print(result.stdout)
    elif result.returncode != 0:
        error_message = (
            f"aton.st.call.bash: Command failed with exit code {result.returncode}.\n"
            f"{result.stderr.strip()}"
        )
        if not return_anyway:
            raise RuntimeError(error_message)
    return result


def git(
        path=None,
        verbose=True,
        message=None,
        tag=None
    ) -> None:
    """Automatically update a Git repository"""
    if path:
        os.chdir(path)
    bash("git fetch", path, verbose)
    bash("git add .", path, verbose)
    if not message:
        date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M")
        message = f'Automatic push on {date} with Aton {__version__}'
    bash(f'git commit -m "{message}"', path, verbose)
    if tag:
        bash(f'git tag -a {tag} HEAD -m {message}', path, verbose)
        bash("git push origin main --tags", path, verbose)
    else:
        bash("git push", path, verbose)
    print("Git updated!")
    return None


def here(folder=None) -> str:
    """Runs the rest of the script inside the specified `folder`.
    
    If none is provided, it runs from the same directory where the current script lies.
    This is really useful to run scripts from the VSCode terminal, etc.
    Returns the path of the used `folder`, or the path of the script if folder is not provided.\n
    Note that this changes not only the working directory of your script,
    but also of other scripts that import and run your script.
    """
    if folder:
        caller = os.path.abspath(folder)
    else:
        caller = os.path.dirname(os.path.abspath(os.path.realpath(sys.argv[0])))
    os.chdir(caller)
    return caller




"""
# Description

This module contains common classes used to load and manipulate spectral data.
Any class can be instantiated directly from the `aton.spx` module,
as `aton.spx.Class()`.


# Index

| | |
| --- | --- |
| `Spectra`  | Used to load and process spectral data |
| `Plotting` | Stores plotting options, used in `Spectra.plotting` |
| `Material` | Used to store and calculate material parameters, such as molar masses and neutron cross sections |


# Examples

To load two INS spectra CSV files with cm$^{-1}$ as input units,
converting them to meV units, and finally plotting them:
```python
from aton import spx
ins = spx.Spectra(
    type     = 'ins',
    files    = ['example_1.csv', 'example_2.csv'],
    units_in = 'cm-1',
    units    = 'meV',
    )
spx.plot(ins)
```

Check more use examples in the [`Aton/examples/`](https://github.com/pablogila/Aton/tree/main/examples) folder.

---
"""


import numpy as np
import pandas as pd
from copy import deepcopy
import os
import aton.st.alias as alias
import aton.phys as phys


class Plotting:
    """Stores plotting options, read by `aton.spx.plot`"""
    def __init__(
            self,
            title:str=None,
            xlim=None,
            ylim=None,
            margins:list=[0,0],
            offset=True,
            scaling:float=1.0,
            vline:list=None,
            vline_error:list=None,
            figsize:tuple=None,
            log_xscale:bool=False,
            show_yticks:bool=False,
            xlabel:str=None,
            ylabel:str=None,
            legend=None,
            legend_title:str=None,
            legend_size='medium',
            legend_loc='best',
            save_as:str=None,
        ):
        """Default values can be overwritten when initializing the Plotting object."""
        self.title = title
        """Title of the plot. Set it to an empty string to remove the title."""
        self.xlim = self._set_limits(xlim)
        """List with the x-limits of the plot, as in `[xlim_low, xlim_top]`."""
        self.ylim = self._set_limits(ylim)
        """List with the y-limits of the plot, as in `[ylim_low, ylim_top]`."""
        self.margins = self._set_limits(margins)
        """List with additional margins at the bottom and top of the plot, as in `[low_margin, top_margin]`."""
        self.offset = offset
        """If `True`, the plots will be separated automatically.

        It can be set to a float, to equally offset the plots by a given value.
        """
        self.scaling = scaling
        "Scaling factor"
        if vline is not None and not isinstance(vline, list):
            vline = [vline]
        self.vline = vline
        """Vertical line/s to plot. Can be an int or float with the x-position, or a list with several ones."""
        if vline_error is not None and not isinstance(vline_error, list):
            vline_error = [vline_error]
        self.vline_error = vline_error
        """Plot a shaded area of the specified width around the vertical lines specified at `vline`.

        It can be an array of the same length as `vline`, or a single value to be applied to all.
        """
        self.figsize = figsize
        """Tuple with the figure size, as in matplotlib."""
        self.log_xscale = log_xscale
        """If true, plot the x-axis in logarithmic scale."""
        self.show_yticks = show_yticks
        """Show or not the yticks on the plot."""
        self.xlabel = xlabel
        """Custom label of the x-axis.

        If `None`, the default label will be used.
        Set to `''` to remove the label of the horizontal axis.
        """
        self.ylabel = ylabel
        """Custom label of the y-axis.
        
        If `None`, the default label will be used.
        Set to `''` to remove the label of the vertical axis.
        """
        if not isinstance(legend, list) and legend is not None and legend != False:
            legend = [legend]
        self.legend = legend
        """Legend of the plot.

        If `None`, the filenames will be used as legend.
        Can be a bool to show or hide the plot legend.
        It can also be an array containing the strings to display;
        in that case, elements set to `False` will not be displayed.
        """
        self.legend_title = legend_title
        """Title of the legend, defaults to `None`."""
        self.legend_size = legend_size
        """Size of the legend, as in matplotlib. Defaults to `'medium'`."""
        self.legend_loc = legend_loc
        """Location of the legend, as in matplotlib. Defaults to `'best'`."""
        self.save_as = save_as
        """Filename to save the plot. None by default."""

    def _set_limits(self, limits) -> list:
        """Set the x and y limits of the plot."""
        if limits is None:
            return [None, None]
        if isinstance(limits, tuple):
            limits = list(limits)
        if isinstance(limits, list):
            if len(limits) == 0:
                return [None, None]
            if len(limits) == 1:
                return [None, limits[0]]
            if len(limits) == 2:
                return limits
            else:
                return limits[:2]
        if isinstance(limits, int) or isinstance(limits, float):
            return [None, limits]
        else:
            raise ValueError(f"Unknown plotting limits: Must be specified as a list of two elements, as [low_limit, high_limit]. Got: {limits}")


class Spectra:
    """Spectra object. Used to load and process spectral data.

    Most functions in the `aton.spx` module receive this object as input.
    """
    def __init__(
            self,
            type:str=None,
            comment:str=None,
            files=None,
            dfs=None,
            units=None,
            units_in=None,
            plotting:Plotting=Plotting(),
        ):
        """All values can be set when initializing the Spectra object."""
        self.type = None
        """Type of the spectra: `'INS'`, `'ATR'`, or `'RAMAN'`."""
        self.comment = comment
        """Custom comment. If `Plotting.title` is None,  it will be the title of the plot."""
        self.files = None
        """List containing the files with the spectral data.

        Loaded automatically to `dfs` with Pandas at initialization.
        In order for Pandas to read the files properly, note that the column lines must start by `#`.
        Any additional line that is not data must be removed or commented with `#`.
        CSV files must be formatted with the first column as the energy or energy transfer,
        and the second column with the intensity or absorbance, depending on the case.
        An additional third `'Error'` column can be used.
        """
        self.dfs = None
        """List containing the pandas dataframes with the spectral data.
        
        Loaded automatically from `files` at initialization.
        """
        self.units = None
        """Target units of the spectral data.
        
        Can be `'meV'` or `'cm-1'`."""
        self.units_in = None
        """Input units of the spectral data, used in the input CSV files.
        
        Can be `'meV'` or `'cm-1'`.
        If the input CSV files have different units,
        it can also be set as a list of the same length of the number of input files,
        eg. `['meV', 'cm-1', 'cm-1']`.
        """
        self.plotting = plotting
        """`Plotting` object, used to set the plotting options."""

        self = self._set_type(type)
        self = self._set_dataframes(files, dfs)
        self = self.set_units(units, units_in)

    def _set_type(self, type):
        """Set and normalize the type of the spectra: `INS`, `ATR`, or `RAMAN`."""
        if type.lower() in alias.experiments['ins']:
            self.type = 'ins'
        elif type.lower() in alias.experiments['atr']:
            self.type = 'atr'
        elif type.lower() in alias.experiments['raman']:
            self.type = 'raman'
        else:
            self.type = type.lower()
        return self

    def _set_dataframes(self, files, dfs):
        '''Set the dfs list of dataframes, from the given files or dfs.'''
        if isinstance(files, list):
            self.files = files
        elif isinstance(files, str):
            self.files = [files]
        else:
            self.files = []

        if isinstance(dfs, pd.DataFrame):
            self.dfs = [dfs]
        elif isinstance(dfs, list) and isinstance(dfs[0], pd.DataFrame):
            self.dfs = dfs
        else:
            self.dfs = [self._read_dataframe(filename) for filename in self.files]
        return self

    def _read_dataframe(self, filename):
        """Read a dataframe from a file."""
        root = os.getcwd()
        file_path = os.path.join(root, filename)
        df = pd.read_csv(file_path, comment='#')
        df = df.sort_values(by=df.columns[0]) # Sort the data by energy

        print(f'\nNew dataframe from {filename}')
        print(df.head(),'\n')
        return df

    def set_units(
            self,
            units,
            units_in=None,
            default_unit='cm-1',
            ):
        """Method to change between spectral units. ALWAYS use this method to do that.

        For example, to change to meV from cm-1:
        ```python
        Spectra.set_units('meV', 'cm-1')
        ```
        """
        mev = 'meV'
        cm = 'cm-1'
        unit_format={
                mev: alias.units['meV'],
                cm: alias.units['cm-1'] + alias.units['cm'],
            }
        if self.units is not None:
            units_in = deepcopy(self.units)
            self.units = units
        elif units is not None:
            units_in = units_in
            self.units = deepcopy(units)
        elif units is None and units_in is None:
            units_in = None
            self.units = default_unit
        elif units is None and units_in is not None:
            units_in = None
            self.units = deepcopy(units_in)
        if isinstance(units_in, list):
            for i, unit_in in enumerate(units_in):
                for key, value in unit_format.items():
                    if unit_in in value:
                        units_in[i] = key
                        break
            if len(units_in) == 1:
                units_in = units_in * len(self.files)
            elif len(units_in) != len(self.files):
                raise ValueError("units_in must be a list of the same length as files.")
        if isinstance(units_in, str):
            for key, value in unit_format.items():
                if units_in in value:
                    units_in = key
                    break
            units_in = [units_in] * len(self.files)
        if isinstance(self.units, list):
            for i, unit in enumerate(self.units):
                for key, value in unit_format.items():
                    if unit in value:
                        self.units[i] = key
                        break
            if len(self.units) == 1:
                self.units = self.units * len(self.files)
            elif len(self.units) != len(self.files):
                raise ValueError("units_in must be a list of the same length as files.")
        if isinstance(self.units, str):
            for key, value in unit_format.items():
                if self.units in value:
                    self.units = key
                    break
            self.units = [self.units] * len(self.files)
        if units_in is None:
            return self
        # Otherwise, convert the dfs
        if len(self.units) != len(units_in):
            raise ValueError("Units len mismatching.")
        for i, unit in enumerate(self.units):
            if unit == units_in[i]:
                continue
            if unit == mev and units_in[i] == cm:
                self.dfs[i][self.dfs[i].columns[0]] = self.dfs[i][self.dfs[i].columns[0]] * phys.cm1_to_meV
            elif unit == cm and units_in[i] == mev:
                self.dfs[i][self.dfs[i].columns[0]] = self.dfs[i][self.dfs[i].columns[0]] * phys.meV_to_cm1
            else:
                raise ValueError(f"Unit conversion error between '{unit}' and '{units_in[i]}'")
        # Rename dataframe columns
        E_units = None
        for i, df in enumerate(self.dfs):
            if self.units[i] == mev:
                E_units = 'meV'
            elif self.units[i] == cm:
                E_units = 'cm-1'
            else:
                E_units = self.units[i]
            if self.type == 'ins':
                if self.dfs[i].shape[1] == 3:
                    self.dfs[i].columns = [f'Energy transfer / {E_units}', 'S(Q,E)', 'Error']
                else:
                    self.dfs[i].columns = [f'Energy transfer / {E_units}', 'S(Q,E)']
            elif self.type == 'atr':
                if self.dfs[i].shape[1] == 3:
                    self.dfs[i].columns = [f'Wavenumber / {E_units}', 'Absorbance', 'Error']
                else:
                    self.dfs[i].columns = [f'Wavenumber / {E_units}', 'Absorbance']
            elif self.type == 'raman':
                if self.dfs[i].shape[1] == 3:
                    self.dfs[i].columns = [f'Raman shift / {E_units}', 'Counts', 'Error']
                else:
                    self.dfs[i].columns = [f'Raman shift / {E_units}', 'Counts']
        return self


class Material:
    """Material class.

    Used to calculate molar masses and cross sections,
    and to pass data to different analysis functions
    such as `aton.spectra.deuterium.impulse_approx().`
    """
    def __init__(
            self,
            elements:dict,
            name:str=None,
            grams:float=None,
            grams_error:float=None,
            mols:float=None,
            mols_error:float=None,
            molar_mass:float=None,
            cross_section:float=None,
            peaks:dict=None,
        ):
        """
        All values can be set when initializing the Material object.
        However, it is recommended to only set the elements and the grams,
        and optionally the name, and calculate the rest with `Material.set()`.
        """
        self.elements = elements
        """Dict of atoms in the material, as in `{'H': 6, 'C':1, 'N':1}`.

        Isotopes can be expressed as 'H2', 'He4', etc. with the atom symbol + isotope mass number.
        """
        self.name = name
        """String with the name of the material."""
        self.grams = grams
        """Mass, in grams."""
        self.grams_error = grams_error
        """Error of the measured mass in grams.

        Set automatically with `Material.set()`.
        """
        self.mols = mols
        """Number of moles.

        Set automatically with `Material.set()`.
        """
        self.mols_error = mols_error
        """Error of the number of moles.

        Set automatically with `Material.set()`.
        """
        self.molar_mass = molar_mass
        """Molar mass of the material, in mol/g.

        Calculated automatically with `Material.set()`.
        """
        self.cross_section = cross_section
        """Neutron total bound scattering cross section, in barns.

        Calculated automatically with `Material.set()`.
        """
        self.peaks = peaks
        """Dict with interesting peaks that you might want to store for later use."""

    def _set_grams_error(self):
        """Set the error in grams, based on the number of decimal places."""
        if self.grams is None:
            return
        decimal_accuracy = len(str(self.grams).split('.')[1])
        # Calculate the error in grams
        self.grams_error = 10**(-decimal_accuracy)

    def _set_mass(self):
        """Set the molar mass of the material.

        If `Material.grams` is provided, the number of moles will be
        calculated and overwritten. Isotopes can be used as 'element + A',
        eg. `'He4'`. This gets splitted with `aton.phys.elements.split_isotope`.
        """
        material_grams_per_mol = 0.0
        for key in self.elements:
            try:
                material_grams_per_mol += self.elements[key] * phys.atoms[key].mass
            except KeyError: # Split the atomic flag as H2, etc
                element, isotope = phys.split_isotope(key)
                material_grams_per_mol += self.elements[key] * phys.atoms[element].isotope[isotope].mass
        self.molar_mass = material_grams_per_mol
        if self.grams is not None:
            self._set_grams_error()
            self.mols = self.grams / material_grams_per_mol
            self.mols_error = self.mols * np.sqrt((self.grams_error / self.grams)**2)
    
    def _set_cross_section(self):
        """Set the cross section of the material, based on the `elements` dict.

        If an isotope is used, eg. `'He4'`, it splits the name with `aton.phys.functions.split_isotope`.
        """
        total_cross_section = 0.0
        for key in self.elements:
            try:
                total_cross_section += self.elements[key] * phys.atoms[key].cross_section
            except KeyError: # Split the atomic flag as H2, etc
                element, isotope_index = phys.split_isotope(key)
                total_cross_section += self.elements[key] * phys.atoms[element].isotope[isotope_index].cross_section
        self.cross_section = total_cross_section

    def set(self):
        """Set the molar mass, cross section and errors of the material."""
        self._set_mass()
        self._set_cross_section()

    def print(self):
        """Print a summary with the material information."""
        print('\nMATERIAL')
        if self.name is not None:
            print(f'Name: {self.name}')
        if self.grams is not None and self.grams_error is not None:
            print(f'Grams: {self.grams} +- {self.grams_error} g')
        elif self.grams is not None:
            print(f'Grams: {self.grams} g')
        if self.mols is not None and self.mols_error is not None:
            print(f'Moles: {self.mols} +- {self.mols_error} mol')
        elif self.mols is not None:
            print(f'Moles: {self.mols} mol')
        if self.molar_mass is not None:
            print(f'Molar mass: {self.molar_mass} g/mol')
        if self.cross_section is not None:
            print(f'Cross section: {self.cross_section} barns')
        if self.elements is not None:
            print(f'Elements: {self.elements}')
        print('')




"""
# Description

Functions to manipulate the content of text files.


# Index

`insert_at()`  
`insert_under()`  
`replace()`  
`replace_line()`  
`replace_between()`  
`delete_under()`  
`correct_with_dict()`  
`from_template()`  

---
"""


import mmap
from . import find
import aton.st.file as file


def insert_at(
        filepath,
        text:str,
        position:int
    ) -> None:
    """Inserts a `text` in the line with `position` index of a given `filepath`.

    If `position` is negative, starts from the end of the file.
    """
    file_path = file.get(filepath)
    with open(file_path, 'r+') as f:
        lines = f.read().splitlines()
        if position < 0:
            position = len(lines) + position + 1
        if position < 0 or position > len(lines):
            raise IndexError("Position out of range")
        lines.insert(position, text)
        f.seek(0)
        f.write('\n'.join(lines))
        f.truncate()
    return None


def insert_under(
        filepath,
        key:str,
        text:str,
        insertions:int=0,
        skips:int=0,
        regex:bool=False
    ) -> None:
    """Inserts a `text` under the line(s) containing the `key` in `filepath`.

    The keyword can be at any position within the line.
    Regular expressions can be used by setting `regex=True`. 

    By default all matches are inserted with `insertions=0`,
    but it can insert only a specific number of matches
    with positive numbers (1, 2...), or starting from the bottom with negative numbers.

    The text can be introduced after a specific number of lines after the match,
    changing the value `skips`. Negative integers introduce the text in the previous lines.
    """
    file_path = file.get(filepath)
    if regex:
        positions = find.pos_regex(file_path, key, insertions)
    else:
        positions = find.pos(file_path, key, insertions)
    positions.reverse()  # Must start replacing from the end, otherwise the atual positions may change!
    # Open the file in read-write mode
    with open(file_path, 'r+b') as f:
        with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_WRITE) as mm:
            # Get the places to insert the text
            for position in positions:
                start, end = find.line_pos(mm, position, skips)
                inserted_text = '\n' + text # Ensure we end in a different line
                if end == 0: # If on the first line
                    inserted_text = text + '\n'
                remaining_lines = mm[end:]
                new_line = inserted_text.encode()
                updated_content = new_line + remaining_lines
                mm.resize(len(mm) + len(new_line))
                mm[end:] = updated_content
    return None


def replace(
        filepath:str,
        key:str,
        text:str,
        replacements:int=0,
        regex:bool=False
    ) -> None:
    """Replaces the `key` with `text` in `filepath`.

    It can also be used to delete the keyword with `text=''`.
    To search with regular expressions, set `regex=True`.

    The value `replacements` specifies the number of replacements to perform:
    1 to replace only the first keyword found, 2, 3...
    Use negative values to replace from the end of the file,
    eg. to replace the last found key, use `replacements=-1`.
    To replace all values, set `replacements = 0`, which is the value by default.

    ```
    line... key ...line -> line... text ...line
    ```
    """
    file_path = file.get(filepath)
    if regex:
        positions = find.pos_regex(file_path, key, replacements)
    else:
        positions = find.pos(file_path, key, replacements)
    positions.reverse()  # Must start replacing from the end, otherwise the atual positions may change!
    with open(file_path, 'r+') as f:
        content = f.read()
        for start, end in positions:
            content = "".join([content[:start], text, content[end:]])
        f.seek(0)
        f.write(content)
        f.truncate()
    return None


def replace_line(
        filepath:str,
        key:str,
        text:str,
        replacements:int=0,
        skips:int=0,
        additional:int=0,
        regex:bool=False
    ) -> None:
    """Replaces the entire line(s) containing the `key` with the `text` in `filepath`.

    It can be used to delete line(s) by setting `text=''`.
    Regular expressions can be used with `regex=True`.

    The value `replacements` specifies the number of lines to replace:
    1 to replace only the first line with the keyword, 2, 3...
    Use negative values to replace from the end of the file,
    e.g., to replace only the last line containing the keyword, use `replacements = -1`.
    To replace all lines, set `replacements = 0`, which is the value by default.

    The default line to replace is the matching line,
    but it can be any other specific line after or before the matching line;
    this is indicated with `skips` as a positive or negative integer.

    More lines can be replaced with `additional` lines (int).
    Note that the matched line plus the additional lines
    will be replaced, this is, additional lines +1.
    """
    file_path = file.get(filepath)
    if regex:
        positions = find.pos_regex(file_path, key, replacements)
    else:
        positions = find.pos(file_path, key, replacements)
    positions.reverse()  # Must start replacing from the end, otherwise the atual positions may change!
    # Open the file in read-write mode
    with open(file_path, 'r+b') as f:
        with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_WRITE) as mm:
            for position in positions:
                # Get the positions of the full line containing the match
                line_start, line_end = find.line_pos(mm, position, skips)
                # Additional lines
                if additional > 0:
                    for _ in range(abs(additional)):
                        line_end = mm.find(b'\n', line_end + 1, len(mm)-1)
                        if line_end == -1:
                            line_end = len(mm) - 1
                            break
                elif additional < 0:
                    for _ in range(abs(additional)):
                        line_start = mm.rfind(b'\n', 0, line_start - 1) + 1
                        if line_start == -1:
                            line_start = 0
                            break
                # Replace the line
                old_line = mm[line_start:line_end]
                new_line = text.encode()
                if text == '':  # Delete the line, and the extra \n
                    remaining_content = mm[line_end:]
                    mm.resize(len(mm) - len(old_line) - 1)
                    mm[line_start-1:] = remaining_content
                # Directly modify the memory-mapped region
                elif len(new_line) == len(old_line):
                    mm[line_start:line_end] = new_line
                else:  # Adjust content for differing line sizes
                    remaining_content = mm[line_end:]
                    updated_content = new_line + remaining_content
                    mm.resize(len(mm) + len(new_line) - len(old_line))
                    mm[line_start:] = updated_content
    return None


def replace_between(
        filepath:str,
        key1:str,
        key2:str,
        text:str,
        delete_keys:bool=False,
        from_end:bool=False,
        regex:bool=False
    ) -> None:
    """Replace with `text` between keywords `key1` and `key2` in `filepath`.

    It can be used to delete the text between the keys by setting `text=''`.
    Regular expressions can be used by setting `regex=True`.

    Key lines are also deleted if `delete_keys=True`.

    Only the first matches of the keywords are used by default;
    you can use the last ones with `from_end = True`.
    ```
    lines...
    key1
    text
    key2
    lines...
    ```
    """
    file_path = file.get(filepath)
    index = 1
    if from_end:
        index = -1
    start, end = find.between_pos(file_path, key1, key2, delete_keys, index, regex)
    with open(file_path, 'r+b') as f:
        with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_WRITE) as mm:
            # Replace the line
            old_content = mm[start:end]
            new_content = text.encode()
            if text == '':  # Delete the line, and the extra \n
                remaining_content = mm[end:]
                mm.resize(len(mm) - len(old_content) - 1)
                mm[start-1:] = remaining_content
            # Directly modify the memory-mapped region
            elif len(new_content) == len(old_content):
                mm[start:end] = new_content
            else:  # Adjust the content for differing line sizes
                remaining_content = mm[end:]
                updated_content = new_content + remaining_content
                mm.resize(len(mm) + len(new_content) - len(old_content))
                mm[start:] = updated_content
    return None


def delete_under(
        filepath,
        key:str,
        match:int=1,
        skips:int=0,
        regex:bool=False
    ) -> None:
    """Deletes all the content under the line containing the `key` in `filepath`.

    The keyword can be at any position within the line.
    Regular expressions can be used by setting `regex=True`.

    By default the first `match` is used; it can be any positive integer (0 is treated as 1!),
    including negative integers to select a match starting from the end of the file.

    The content can be deleted after a specific number of lines after the match,
    changing the value `skips`, that skips the specified number of lines.
    Negative integers start deleting the content from the previous lines.
    """
    file_path = file.get(filepath)
    if match == 0:
        match = 1
    if regex:
        positions = find.pos_regex(file_path, key, match)
    else:
        positions = find.pos(file_path, key, match)
    if match > 0:  # We only want one match, and should be the last if matches > 0
        positions.reverse()
    position = positions[0]
    # Open the file in read-write mode
    with open(file_path, 'r+b') as f:
        with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_WRITE) as mm:
            # Get the places to insert the text
            start, end = find.line_pos(mm, position, skips)
            mm.resize(len(mm) - len(mm[end:]))
            mm[end:] = b''
    return None


def correct_with_dict(
        filepath:str,
        correct:dict
    ) -> None:
    """Corrects the given text file `filepath` using a `correct` dictionary."""
    file_path = file.get(filepath)
    with open(file_path, 'r+') as f:
        content = f.read()
        for key, value in correct.items():
            content = content.replace(key, value)
        f.seek(0)
        f.write(content)
        f.truncate()
    return None


def from_template(
        old:str,
        new:str,
        correct:dict=None,
        comment:str=None,
    ) -> None:
    """Creates `new` file from `old`, replacing values from a `correct` dict, inserting a `comment` on top."""
    file.copy(old, new)
    if comment:
        insert_at(new, comment, 0)
    if correct:
        correct_with_dict(new, correct)
    return None




'''
# Description

This module contains functions to sort and analyse
chemical data using the `aton.phys.atoms` megadictionary,
which contains the properties of all elements.
It also contains the tools needed to
automatically update said megadictionary.


# Index

| | |
| --- | --- |
| `export_atoms()`     | Used to update and export the `aton.phys.atoms` dict |
| `split_isotope()`    | Splits element name and mass number |
| `allowed_isotopes()` | Returns the available mass numbers for a given element |


# Examples

All functions can be called from the phys subpackage directly, as:
```python
from aton import phys
phys.split_isotope('He4')    # (He, 4)
phys.allowed_isotopes('Li')  # (6, 7)
```

---
'''


def export_atoms(
        atoms:dict,
        filename='exported_atoms.py'
    ) -> None:
    '''Export a dictionary of chemical elements to a python file.

    This is used to build and update the `aton.atoms` megadictionary, that contains
    all the element data, such as masses, cross-sections, etc.
    The `atoms.py` file must be modified here.
    '''
    with open(filename, 'w') as f:
        # Write the docstrings
        f.write(
            "'''\n"
            "# Description\n\n"
            "This module contains the `atoms` megadictionary,\n"
            "which contains the properties of all elements.\n"
            "It is managed and updated automatically with `aton.phys.functions`.\n\n\n"
            "# Index\n\n"
            "| | |\n"
            "| --- | --- |\n"
            "| `Element` | Used as values in the `atoms` dict, stores element properties |\n"
            "| `Isotope` | Used as values in `Element.isotope`, stores isotope properties |\n"
            "| `atoms`   | The dict with data from all elements |\n\n\n"
            "# Examples\n\n"
            "The `atoms` dictionary can be loaded directly as `aton.phys.atoms`, as shown below:\n"
            "```python\n"
            "from aton import phys\n"
            "aluminium_cross_section = phys.atoms['Al'].cross_section\n"
            "He4_mass = phys.atoms['H'].isotope[4].mass\n\n"
            "```\n\n"
            "---\n"
            "'''\n\n\n"
            "class Element:\n"
            "    '''Used in the `aton.atoms` megadictionary to store element data.'''\n"
            "    def __init__(self=None, Z:int=None, symbol:str=None, name:str=None, mass:float=None, cross_section:float=None, isotope:dict=None):\n"
            "        self.Z: int = Z\n"
            "        '''Atomic number (Z). Corresponds to the number of protons / electrons.'''\n"
            "        self.symbol: str = symbol\n"
            "        '''Standard symbol of the element.'''\n"
            "        self.name: str = name\n"
            "        '''Full name.'''\n"
            "        self.mass: float = mass\n"
            "        '''Atomic mass, in atomic mass units (amu).'''\n"
            "        self.cross_section: float = cross_section\n"
            "        '''Total bound scattering cross section.'''\n"
            "        self.isotope: dict = isotope\n"
            "        '''\n"
            "        Dictionary containing the different `Isotope` of the element.\n"
            "        The keys are the mass number (A).\n"
            "        '''\n\n\n"
            "class Isotope:\n"
            "    '''Used in the `aton.atoms` megadictionary to store isotope data.'''\n"
            "    def __init__(self, A:int=None, mass:float=None, abundance:float=None, cross_section:float=None):\n"
            "        self.A: int = A\n"
            "        '''\n"
            "        Mass number (A) of the isotope.\n"
            "        Corresponds to the total number of protons + neutrons in the core.\n"
            "        '''\n"
            "        self.mass: float = mass\n"
            "        '''Atomic mass of the isotope, in atomic mass units (amu).'''\n"
            "        self.abundance: float = abundance\n"
            "        '''Relative abundance of the isotope.'''\n"
            "        self.cross_section: float = cross_section\n"
            "        '''Total bound scattering cross section of the isotope.'''\n\n\n"
        )
        # Start the atom megadictionary
        f.write("atoms = {\n")
        for key, element in atoms.items():
            f.write(f"    '{element.symbol}': Element(\n"
                    f"        Z             = {element.Z},\n"
                    f"        symbol        = '{element.symbol}',\n"
                    f"        name          = '{element.name}',\n")
            if element.mass:
                f.write(f"        mass          = {element.mass},\n")
            if element.cross_section:
                f.write(f"        cross_section = {element.cross_section},\n")
            if element.isotope:
                f.write("        isotope       = {\n")
                for iso in element.isotope.values():
                    f.write(f"            {iso.mass_number} : Isotope(\n")
                    if iso.mass_number:
                        f.write(f"                A             = {iso.A},\n")
                    if iso.mass:
                        f.write(f"                mass          = {iso.mass},\n")
                    if iso.abundance:
                        f.write(f"                abundance     = {iso.abundance},\n")
                    if iso.cross_section:
                        f.write(f"                cross_section = {iso.cross_section},\n")
                    f.write(f"                ),\n")
                f.write("            },\n")
            f.write(f"        ),\n")
        f.write("}\n")
        print(f'Exported elements to {filename}')
    return None


def split_isotope(name:str) -> tuple:
    """Split the `name` of an isotope into the element and the mass number, eg. He4 -> He, 4.

    If the isotope is not found in the `aton.atoms` megadictionary it raises an error,
    informing of the allowed mass numbers (A) values for the given element.
    """
    element = ''.join(filter(str.isalpha, name))
    isotope = int(''.join(filter(str.isdigit, name)))
    isotopes = allowed_isotopes(element)
    if not isotope in isotopes:
        raise KeyError(f'Unrecognised isotope: {name}. Allowed mass numbers for {element} are: {isotopes}')
    return element, isotope


def allowed_isotopes(element) -> list:
    '''Return a list with the allowed mass numbers (A) of a given `element`.

    These mass numbers are used as isotope keys in the `aton.atoms` megadictionary.
    '''
    from .atoms import atoms
    if not element in atoms.keys():
        try:
            element, _ = split_isotope(element)
        except KeyError:
            raise KeyError(f'Unrecognised element: {element}')
    isotopes = atoms[element].isotope.keys()
    return isotopes




'''
# Description

This module contains useful constants and conversion factors.


# Index

[Energy conversion factors](#energy-conversion-factors)  
[Distance conversion factors](#distance-conversion-factors)  
[Mass conversion factors](#mass-conversion-factors)  
[Pressure conversion factors](#pressure-conversion-factors)  
[Time conversion factors](#time-conversion-factors)  
[Universal constants](#universal-constants)  


# Examples

```python
from aton import phys
energy_in_cm1 = 1000 * phys.meV_to_cm1       # 8065.5
length_in_angstroms = 10.0 * phys.bohr_to_A  # 5.29177210544
phys.hbar   # 1.0545718176461565e-34
```

---

## Energy conversion factors
Note that `cm` refers to cm$^{-1}$.
'''

import numpy as np

eV_to_meV   = 1000.0
meV_to_eV   = 0.001
meV_to_cm1  = 8.0655
cm1_to_meV  = 1.0 / meV_to_cm1
eV_to_J     = 1.602176634e-19
J_to_eV     = 1.0 / eV_to_J
meV_to_J    = meV_to_eV * eV_to_J
J_to_meV    = J_to_eV * eV_to_meV
Ry_to_eV    = 13.605693122990
eV_to_Ry    = 1.0 / Ry_to_eV
Ry_to_J     = 2.1798723611030e-18
J_to_Ry     = 1.0 / Ry_to_J
cal_to_J    = 4.184
J_to_cal    = 1 / cal_to_J
kcal_to_J   = cal_to_J * 1000.0
J_to_kcal   = 1 / kcal_to_J

'''---
## Distance conversion factors
Note that `A` refers to Angstroms.
'''
A_to_m      = 1.0e-10
m_to_A      = 1.0 / A_to_m
bohr_to_m   = 5.29177210544e-11
m_to_bohr   = 1.0 / bohr_to_m
A_to_bohr   = A_to_m * m_to_bohr
bohr_to_A   = 1.0 / A_to_bohr

'''---
## Mass conversion factors
'''
amu_to_kg   = 1.66053906660e-27
kg_to_amu   = 1.0 / amu_to_kg
kg_to_g     = 1000.0
g_to_kg     = 1.0 / kg_to_g

'''---
## Pressure conversion factors
'''
GPa_to_Pa   = 1.0e9
Pa_to_GPa   = 1.0 / GPa_to_Pa
kbar_to_bar = 1000.0
bar_to_kbar = 1.0 / kbar_to_bar
Pa_to_bar   = 1.0e-5
bar_to_Pa   = 1.0 / Pa_to_bar
GPa_to_kbar = GPa_to_Pa * Pa_to_bar * bar_to_kbar
kbar_to_GPa = 1.0 / GPa_to_kbar

'''---
## Time conversion factors
Note that `H` refers to hours.
'''
H_to_s      = 3600.0
s_to_H      = 1.0 / H_to_s

'''---
## Universal constants
Given in SI units unless stated otherwise.
'''
h = 6.62607015e-34      # J s
'''Planck constant, in J·s.'''
h_eV = h * J_to_eV
'''Planck constant, in eV·s.'''
hbar = h / (2 * np.pi)  # J s
'''Reduced Planck constant, in J·s.'''
hbar_eV = h_eV / (2 * np.pi)
'''
Reduced Planck constant, in eV·s.
'''




"""
# Description

Functions to work with [CASTEP](https://castep-docs.github.io/castep-docs/) calculation files.

# Index

| | |
| --- | --- |
| `read_castep()` | Output reading |

---
"""


import aton.st.file as file
import aton.text.find as find
import aton.text.extract as extract


def read_castep(filename) -> dict:
    """
    Reads a CASTEP output file, specified in `filename`.
    Returns a dictionary with the following keys:
    `'Enthalpy'` (LBFGS: Final Enthalpy, in kJ/mol),
    `'Energy'` (Total energy corrected for finite basis set, in eV),
    `'Space group'`, `'Volume'` (Angstrom^3), `'Density'` (amu/Angstrom^3), `'Density_g'` (g/cm^3),
    `'A'`, `'B'`, `'C'` (Angstroms), `'Alpha'`, `'Beta'`, `'Gamma'` (Degrees).\n
    Note that these output keys start with a **C**apital letter.
    """
    file_castep = file.get(filename)
    # Initial definitions
    enthalpy    = None
    energy      = None
    space_group = None
    volume      = None
    density     = None
    density_g   = None
    a           = None
    b           = None
    c           = None
    alpha       = None
    beta        = None
    gamma       = None
    # Find the output values in the file
    enthalpy_str    = find.lines(file_castep, 'LBFGS: Final Enthalpy     =', -1)
    energy_str      = find.lines(file_castep, 'Total energy corrected for finite basis set =', -1)
    space_group_str = find.lines(file_castep, 'Space group of crystal =', -1)
    volume_str      = find.lines(file_castep, 'Current cell volume =', -1)
    density_str     = find.lines(file_castep, 'density =', -1, 1)
    a_str           = find.lines(file_castep, 'a =', -1)
    b_str           = find.lines(file_castep, 'b =', -1)
    c_str           = find.lines(file_castep, 'c =', -1)
    
    if enthalpy_str:
        enthalpy = extract.number(enthalpy_str[0], 'LBFGS: Final Enthalpy')
    if energy_str:
        energy = extract.number(energy_str[0], 'Total energy corrected for finite basis set')
    if space_group_str:
        # Avoid little stupid errors
        space_group_str = space_group_str.replace(',','.')
        space_group = extract.string(space_group_str[0], 'Space group of crystal')
    if volume_str:
        volume = extract.number(volume_str[0], 'Current cell volume')
    if density_str:
        density = extract.number(density_str[0], 'density')
        density_g = extract.number(density_str[1], '')
    if a:
        a = extract.number(a_str, 'a')
        alpha = extract.number(a_str, 'alpha')
    if b:
        b = extract.number(b_str, 'b')
        beta = extract.number(b_str, 'beta')
    if c:
        c = extract.number(c_str, 'c')
        gamma = extract.number(c_str, 'gamma')
    # Return as a dict
    dictionary = {
        'Enthalpy'    : enthalpy,
        'Energy'      : energy,
        'Space group' : space_group,
        'Volume'      : volume,
        'Density'     : density,
        'Density_g'   : density_g,
        'A'           : a,
        'B'           : b,
        'C'           : c,
        'Alpha'       : alpha,
        'Beta'        : beta,
        'Gamma'       : gamma,
    }
    return dictionary




'''
# Description

Functions to extract data from raw text strings.


# Index

`number()`  
`string()`  
`column()`  
`coords()`  
`element()`  


# Examples

To extract a float value from a string,
```python
from aton import text
line = 'energy =   500.0 Ry'
text.extract.number(line, 'energy')
>>> 500.0  # float output
```

To extract a text value, after and before specific strings,
```python
line = 'energy =   500.0 Ry were calculated'
text.extract.string(line, 'energy', 'were')
>>> '500.0 Ry'  # String output
```

To extract a float value from a specific column,
```python
# Name, Energy, Force, Error
line = 'Testing    1.1    1.2    0.3'
energy = text.extract.column(line, 1)
>>> 1.1
```

To extract coordinates,
```python
line = ' He  0.10  0.20  0.30 '
text.extract.coords(line)
>>> [0.1, 0.2, 0.3]  # List of floats
```

To extract chemical elements,
```python
line = ' He4  0.10  Ag  0.20  Pb  0.30 '
first_element = text.extract.element(line, 0)
>>> 'He4'
third_element = text.extract.element(line, 2)
>>> 'Pb'
```

---
'''


import re
import aton.phys as phys


def number(
        text:str,
        name:str=''
    ) -> float:
    """Extracts the float value of a given `name` variable from a raw `text`."""
    if text == None:
        return None
    pattern = re.compile(rf"{name}\s*[:=]?\s*(-?\d+(?:\.\d+)?(?:[eEdD][+\-]?\d+)?)")
    match = pattern.search(text)
    if match:
        return float(match.group(1))
    return None
    

def string(
        text:str,
        name:str='',
        stop:str='',
        strip:bool=True
    ) -> str:
    """Extracts the `text` value of a given `name` variable from a raw string. Stops before an optional `stop` string.

    Removes leading and trailing commas by default, change this with `strip=False`.
    """
    pattern = re.compile(rf"{name}\s*[:=]?\s*(.*)")
    if stop:
        pattern = re.compile(rf"{name}\s*[:=]?\s*(.*)(?={stop})")
    match = re.search(pattern, text)
    if not match:
        return None
    result = str(match.group(1))
    result = result.strip()
    if strip:
        result = result.strip("'")
        result = result.strip('"')
        result = result.strip()
    return result


def column(
        text:str,
        column:int=0
    ) -> float:
    """Extracts the desired float `column` index of a given `string` (0 by default)."""
    if text is None:
        return None
    columns = text.split()
    pattern = r'(-?\d+(?:\.\d+)?(?:[eE][+\-]?\d+)?)'
    if column < len(columns):
        match = re.match(pattern, columns[column])
        if match:
            return float(match.group(1))
    return None


def coords(text:str) -> list:
    """Returns a list with the float coordinates expressed in a given `text` string."""
    if text is None:
        return None
    columns = re.split(r'[,\s]+', text.strip())
    pattern = r'(-?\d+(?:\.\d+)?(?:[eE][+\-]?\d+)?)'
    matches = []
    for column in columns:
        match = re.match(pattern, column)
        if match:
            matches.append(float(match.group(1)))
    return matches


def element(text:str, index:int=0) -> str:
    """Extract a chemical element from a raw `text` string.

    If there are several elements, you can return a specific `index` match (positive, 0 by default).
    Allows for standard elements (H, He, Na...) and isotopes (H2, He4...).
    """
    if text is None:
        return None
    columns = re.split(r'[,\s]+', text.strip())
    pattern = r'\s*([A-Z][a-z]{0,2}\d{0,3})(?=\s|$)'
    matches = []
    for column in columns:
        match = re.match(pattern, column)
        if match:
            matches.append(str(match.group(1)))
    # We have a list with possible matches. Let's determine which are actual elements.
    found_elements = []
    for possible_element in matches:
        possible_element = possible_element.strip()
        if not possible_element in phys.atoms.keys():
            try:
                element, isotope = phys.split_isotope(possible_element)
            except:  # It is not a valid atom
                continue
        found_elements.append(possible_element)
    if len(found_elements) <= index:
        return found_elements[-1]
    return found_elements[index]




from aton.text import extract


def test_extract_number():
    assert extract.number(' test = 123 number ', 'test') == 123.0
    assert extract.number(' test 123 number ', 'test') == 123.0


def test_extract_string():
    assert extract.string(text=' test = "hello" stop ', name='test', stop='stop', strip=True) == 'hello'
    assert extract.string(text=' test "hello" stop ', name='test', stop='stop', strip=True) == 'hello'
    assert extract.string(text=" test 'hello' stop ", name='test', stop='stop', strip=True) == 'hello'
    assert extract.string(text=" test 'hello' stop ", name='test', stop='stop', strip=False) == "'hello'"


def test_extract_column():
    assert extract.column(' 123 456.5 789  ', 2) == 789


def test_extract_coords():
    assert extract.coords('coordinates: 1.0, 2.0 and 3 these were the coordinates') ==[1.0, 2.0, 3.0]


def test_extract_element():
    string = '  element I Lead Pb Nitrogen H2, Xx2 fake element, O Oxygen, He4 isotope Ag Element '
    assert extract.element(text=string, index=0) == 'I'
    assert extract.element(text=string, index=1) == 'Pb'
    assert extract.element(text=string, index=2) == 'H2'
    assert extract.element(text=string, index=3) == 'O'
    assert extract.element(text=string, index=4) == 'He4'
    assert extract.element(text=string, index=5) == 'Ag'




"""
# *Ab-initio* interfaces

This module contains interfaces for several *ab-initio* calculation softwares.
These interfaces can be easily expanded with the `aton.text` module.


# Index

| | |  
| --- | --- |  
| `aton.interface.qe`      | Interface for [Quantum ESPRESSO](https://www.quantum-espresso.org/)'s [pw.x](https://www.quantum-espresso.org/Doc/INPUT_PW.html) module |  
| `aton.interface.phonopy` | Interface for [Phonopy](https://phonopy.github.io/phonopy/) calculations |  
| `aton.interface.castep`  | Interface for [CASTEP](https://castep-docs.github.io/castep-docs/) calculations |  


# Examples

## Quantum ESPRESSO

To read the output from a Quantum ESPRESSO pw.x calculation,  
```python
from aton.interface import qe
calculation = qe.read_out('relax.out')  # Read to a dictionary
calculation.keys()                       # See the available values
energy = calculation['Energy']           # Final energy from the calculation
```

To modify values from an input file,  
```python
from aton.interface import qe
qe.add_atom('H  0.10  0.20  0.30')         # Add a hydrogen atom to a specific position
qe.set_value('relax.in', 'ecutwfc', 60.0)  # Set the input ecutwfc value
```

Check the full `aton.interface.qe` API reference for more details.


## Phonopy

To perform a phonon calculation from a relaxed structure via Quantum ESPRESSO,  
```python
from aton import interface
# Create the supercell inputs
interface.phonopy.make(dimension='2 2 2', folder='./calculation')
# Sbatch to a cluster
interface.phonopy.sbatch('./calculation')
```

Check the full `aton.interface.phonopy` API reference for more details.


## CASTEP

To read output values from a CASTEP calculation,  
```python
from aton.interface import castep
output = castep.read_castep('calculation.castep')  # Read the output
energy = output['Energy']                          # Get the final energy
```

Check the full `aton.interface.castep` API reference for more details.

"""


from . import qe
from . import phonopy
from . import castep




"""
# Description

Functions to work with [Phonopy](https://phonopy.github.io/phonopy/) calculations,
along with [Quantum ESPRESSO](https://www.quantum-espresso.org/).  


# Index

Create supercell inputs and sbatch phonon calculations
| | |
| --- | --- |
| `make()`   | Build the supercell inputs |
| `sbatch()` | Sbatch'es the supercell calculations |

For more control  
`supercells_from_scf()`  
`scf_header_to_supercells()`  
`check_slurm_template()`  

---
"""


import os
import re
from aton._version import __version__
import aton.st.file as file
import aton.st.call as call
import aton.text.find as find
import aton.text.edit as edit # text
import aton.text.extract as extract
import aton.interface.qe as qe


def make(
        dimension:str='2 2 2',
        folder:str=None,
        relax_in:str='relax.in',
        relax_out:str='relax.out',
        slurm_template:str='scf.slurm'
    ) -> None:
    '''
    Starting on a given `folder` (CWD if none) from the `relax_in` and `relax_out` (default ones),
    creates the supercells of a `dimension` (`2 2 2` by default)
    needed for the Phonopy calculations with Quantum ESPRESSO.
    It runs sequentially `thotpy.qe.scf_from_relax()`, `supercells_from_scf()` and `scf_header_to_supercells()`.
    Finally, it checks the `slurm_template` with `check_slurm_template()`.
    '''
    print(f'\nWelcome to thotpy.phonopy {__version__}\n'
          'Creating all supercell inputs with Phonopy for Quantum ESPRESSO...\n')
    qe.scf_from_relax(folder, relax_in, relax_out)
    supercells_from_scf(dimension, folder)
    scf_header_to_supercells(folder)
    print('\n------------------------------------------------------\n'
          'PLEASE CHECH BELOW THE CONTENT OF THE supercell-001.in\n'
          '------------------------------------------------------\n')
    call.bash('head -n 100 supercell-001.in')
    print('\n------------------------------------------------------\n'
          'PLEASE CHECH THE CONTENT OF THE supercell-001.in\n'
          'The first 100 lines of the input were printed above!\n'
          '------------------------------------------------------\n\n'
          'If it seems correct, run the calculations with thotpy.phonopy.sbatch()\n')
    check_slurm_template(folder, slurm_template)
    return None


def sbatch(
        folder=None,
        slurm_template:str='scf.slurm',
        testing:bool=False
    ) -> None:
    '''
    Launch all your supercell calculations to a cluster using a SLURM manager.
    Runs from a `folder` (CWD if empty), using a `slurm_template` (`scf.slurm` by default).\n
    If `testing=True` it skips the final sbatching, just printing the commands on the screen.\n
    The slurm template must contain the keywords
    `INPUT_FILE`, `OUTPUT_FILE`, and `JOB_NAME` in the following lines:
    ```
    #SBATCH --job-name=JOB_NAME
    mpirun pw.x -inp INPUT_FILE > OUTPUT_FILE
    ```
    '''
    print(f'\naton.interface.phonopy {__version__}\n'
          'Sbatching all supercells...\n')
    key_input = 'INPUT_FILE'
    key_output = 'OUTPUT_FILE'
    key_jobname = 'JOB_NAME'
    id_pattern = re.compile(r'supercell-(\d\d\d).in')
    slurm_folder = 'slurms'
    folder = call.here(folder)
    # Get supercells and abort if not found
    supercells = file.get_list(folder, 'supercell-')
    if len(supercells) == 0:
        raise FileNotFoundError('Supercells were not found! Did you run thotpy.phonopy.make() ?')
    call.bash(f"mkdir {slurm_folder}", folder, True, True)
    # Get the template
    slurm_file = check_slurm_template(folder, slurm_template)
    if not slurm_file:
        print(f'Aborting... Please correct {slurm_template}\n')
        return None
    for supercell in supercells:
        # Get the file ID
        match = re.search(id_pattern, supercell)
        calc_id = match.group(1)
        # Create slurm file for this supercell
        slurm_id = 'scf-' + calc_id + '.slurm'
        supercell = os.path.basename(supercell)
        supercell_out = supercell.replace('.in', '.out')
        fixing_dict = {
            key_jobname: calc_id,
            key_input: supercell,
            key_output: supercell_out
        }
        edit.from_template(slurm_file, slurm_id, fixing_dict)
        if testing:
            call.bash(f"echo {slurm_id}", folder)
        else:
            call.bash(f"sbatch {slurm_id}", folder, True, False)
        call.bash(f"mv {slurm_id} {slurm_folder}", folder, False, True)  # Do not raise error if we can't move the file
    print(f'\nDone! Temporary slurm files were moved to /{slurm_folder}/\n')


def supercells_from_scf(
        dimension:str='2 2 2',
        folder:str=None,
        scf:str='scf.in'
    ) -> None:
    '''
    Creates supercells of a given `dimension` (`2 2 2` by default) inside a `folder`,
    from a Quantum ESPRESSO `scf` input (`scf.in` by default).
    '''
    print(f'\naton.interface.phonopy {__version__}\n')
    folder = call.here(folder)
    scf_in = file.get(folder, scf, True)
    if scf_in is None:
        raise ValueError('No SCF input found in path!')
    call.bash(f'phonopy --qe -d --dim="{dimension}" -c {scf_in}')
    return None


def scf_header_to_supercells(
        folder:str=None,
        scf:str='scf.in',
    ) -> None:
    '''
    Paste the header from the `scf` file in `folder` to the supercells created by Phonopy.
    '''
    print(f'\naton.interface.phonopy {__version__}\n'
          f'Adding headers to Phonopy supercells for Quantum ESPRESSO...\n')
    folder = call.here(folder)
    # Check if the header file, the scf.in, exists
    scf_file = file.get(folder, scf, True)
    if scf_file is None:
        raise ValueError('No header file found in path!')
    # Check if the supercells exist
    supercells = file.get_list(folder, 'supercell-')
    if supercells is None:
        raise ValueError('No supercells found in path!')
    # Check if the supercells contains '&CONTROL' and abort if so
    supercell_sample = supercells[0]
    is_control = find.lines(supercell_sample, r'(&CONTROL|&control)', 1, 0, False, True)
    if is_control:
        raise ValueError('Supercells already contain &CONTROL! Did you do this already?')
    # Check if the keyword is in the scf file
    is_header = find.lines(scf_file, r'ATOMIC_SPECIES', 1, 0, False, False)
    if not is_header:
        raise ValueError('No ATOMIC_SPECIES found in header!')
    # Copy the scf to a temp file
    temp_scf = '_scf_temp.in'
    file.copy(scf_file, temp_scf)
    # Remove the top content from the temp file
    edit.delete_under(temp_scf, 'K_POINTS', -1, 2, False)
    # Find the new number of atoms and replace the line
    updated_values = find.lines(supercell_sample, 'ibrav', 1)  # !    ibrav = 0, nat = 384, ntyp = 5
    if not updated_values:
        print("!!! Okay listen, this is weird. This code should never be running, "
              "but for some reson we couldn't find the updated values in the supercells. "
              "Please, introduce the NEW NUMBER OF ATOMS in the supercells manually (int):")
        nat = int(input('nat = '))
    else:
        nat = extract.number(updated_values[0], 'nat')
    qe.set_value(temp_scf, 'nat', nat)
    # Remove the lattice parameters, since Phonopy already indicates units
    qe.set_value(temp_scf, 'celldm(1)', '')
    qe.set_value(temp_scf, 'A', '')
    qe.set_value(temp_scf, 'B', '')
    qe.set_value(temp_scf, 'C', '')
    qe.set_value(temp_scf, 'cosAB', '')
    qe.set_value(temp_scf, 'cosAC', '')
    qe.set_value(temp_scf, 'cosBC', '')
    # Add the header to the supercells
    with open(temp_scf, 'r') as f:
        header = f.read()
    for supercell in supercells:
        edit.insert_at(supercell, header, 0)
    # Remove the temp file
    os.remove('_scf_temp.in')
    print('Done!')
    return None


def check_slurm_template(
        folder=None,
        slurm_template:str='scf.slurm'
    ) -> str:
    '''
    Check a `slurm_template` inside `folder`.
    The current working directory is used if `folder` is not provided.
    If the file does not exist or is invalid, creates a `scf_EXAMPLE.slurm` file for reference.
    '''
    folder = call.here(folder)
    slurm_example = 'scf_EXAMPLE.slurm'
    new_slurm_file = os.path.join(folder, slurm_example)
    # Default slurm template
    content =f'''# Automatic slurm template created with aton.interface.phonopy {__version__}. https://github.com/pablogila/ThotPy
#!/bin/bash
#SBATCH --partition=general
#SBATCH --qos=regular
#SBATCH --job-name=JOB_NAME
#SBATCH --ntasks=32
#SBATCH --time=1-00:00:00
#SBATCH --mem=128G
# #SBATCH --mail-user=YOUR@EMAIL
# #SBATCH --mail-type=END

module purge
module load QuantumESPRESSO/7.3-foss-2023a

mpirun pw.x -inp INPUT_FILE > OUTPUT_FILE
'''
    # If the slurm template does not exist, create one
    slurm_file = file.get(folder, slurm_template, True)
    if not slurm_file:
        with open(new_slurm_file, 'w') as f:
            f.write(content)
        print(f'!!! WARNING:  Slurm template missing, so an example was generated automatically:\n'
              f'{slurm_example}\n'
              f'PLEASE CHECK it, UPDATE it and RENAME it to {slurm_template}\n'
              'before running thotpy.phonopy.sbatch()\n')
        return None
    # Check that the slurm file contains the INPUT_FILE, OUTPUT_FILE and JOB_NAME keywords
    key_input = find.lines(slurm_file, 'INPUT_FILE')
    key_output = find.lines(slurm_file, 'OUTPUT_FILE')
    key_jobname = find.lines(slurm_file, 'JOB_NAME')
    missing = []
    if not key_input:
        missing.append('INPUT_FILE')
    if not key_output:
        missing.append('OUTPUT_FILE')
    if not key_jobname:
        missing.append('JOB_NAME')
    if len(missing) > 0:
        with open(new_slurm_file, 'w') as f:
            f.write(content)
        print('!!! WARNING:  Some keywords were missing from your slurm template,\n'
              f'PLEASE CHECK the example at {slurm_example}\n'
              'before running thotpy.phonopy.sbatch()\n'
              f'The following keywords were missing from your {slurm_template}:')
        for key in missing:
            print(key)
        print('')
        return None
    print(f"Your slurm template {slurm_template} SEEMS OKAY, "
          "but don't forget to check it before running thotpy.phonopy.sbatch()\n")
    return slurm_file  # Ready to use!




"""
# Description

This module contains functions to normalize data and other variables.


# Index

| | |
| --- | --- |
| `height()`   | Normalize a `spectra` by height |
| `area()`     | Normalize a `spectra` by the area under the datasets |
| `unit_str()` | Normalize a `unit` string from user input |

---
"""


import aton.st.alias as alias
from .classes import *
from .fit import *


def height(
        spectra:Spectra,
        range:list=None,
        axis:str='x',
        df_index:int=0,
        ) -> Spectra:
    """Normalize a set of `spectra` by height.

    By default it normalises the spectra over the entire range.
    This can be modified by setting a specific range,
    as in `range = [x_min, x_max]` with `axis = 'x'`.
    It can also normalise over manual y-positions,
    for example for peaks with different baselines.
    This can be done by settingch `axis='y'`, and
    `range = [[y_min_1, y_max_1], ..., [y_min_N, y_max_N]]`.

    Heights are normalised with respect to the
    reference dataframe `df_index`, the first one by default.
    """
    sdata = deepcopy(spectra)
    if axis.lower() in alias.spatial['y']:
        return _height_y(sdata, range, df_index)
    df0 = sdata.dfs[df_index]
    if range:
        if not isinstance(range, list):
            raise ValueError("range must be a list")
        if len(range) != 2:
            raise ValueError(f"With axis='x', range must be [xmin, xmax]. Yours was:\n{range}")
        range.sort()
        xmin = range[0]
        xmax = range[1]
    else:
        xmin = min(df0[df0.columns[0]])
        xmax = max(df0[df0.columns[0]])
    df0 = df0[(df0[df0.columns[0]] >= xmin) & (df0[df0.columns[0]] <= xmax)]
    ymax_on_range = df0[df0.columns[1]].max()
    normalized_dataframes = []
    for df in sdata.dfs:
        df_range = df[(df[df.columns[0]] >= xmin) & (df[df.columns[0]] <= xmax)]
        i_ymax_on_range = df_range[df_range.columns[1]].max()
        df[df.columns[1]] =  df[df.columns[1]] * ymax_on_range / i_ymax_on_range
        normalized_dataframes.append(df)
    sdata.dfs = normalized_dataframes
    return sdata


def _height_y(
        sdata:Spectra,
        range:list,
        df_index:int=0,
        ) -> Spectra:
    """Private function to handle normalisation in the y-axis"""
    if not range:
        raise ValueError(f"A range must be specified to normalise the Y axis, as in range=[[y_min_1,y_max_1],...,[y_min_N,y_max_N]]\nYours was:\n{range}")
    if not len(range) == len(sdata.dfs):
        raise ValueError("len(range) must match len(Spectra.dfs) for axis='y'")
    ymax = []
    ymin = []
    for values in range:
        if not isinstance(values, list):
            raise ValueError(f"The range for axis='y' must be a list of lists,\nas in range=[[y_min_1,y_max_1],...,[y_min_N,y_max_N]].\nYours was:\n{range}")
        if len(values) != 2:
            raise ValueError(f"2 values needed to normalise the y-axis, ymin and ymax,\nas in range=[[y_min_1,y_max_1],...,[y_min_N,y_max_N]].\nYours was:\n{range}")
        values.sort()
        i_ymin = values[0]
        i_ymax = values[1]
        ymin.append(i_ymin)
        ymax.append(i_ymax)
    reference_height = ymax[df_index] - ymin[df_index]
    normalized_dataframes = []
    for i, df in enumerate(sdata.dfs):
        height = ymax[i] - ymin[i]
        df[df.columns[1]] =  df[df.columns[1]] * reference_height / height
        normalized_dataframes.append(df)
    sdata.dfs = normalized_dataframes
    return sdata


def area(
        spectra:Spectra,
        range:list=None,
        df_index:int=0
        ) -> Spectra:
    """Normalize `spectra` by the area under the datasets."""
    sdata = deepcopy(spectra)
    df0 = sdata.dfs[df_index]
    if range:
        if len(range) != 2:
            raise ValueError(f"The range must be a list of 2 elements, as in [xmin, xmax]. Yours was:\n{range}")
        range.sort()
        xmin = range[0]
        xmax = range[1]
    else:
        xmin = min(df0[df0.columns[0]])
        xmax = max(df0[df0.columns[0]])
    df0 = df0[(df0[df0.columns[0]] >= xmin) & (df0[df0.columns[0]] <= xmax)]
    area_df0, _ = area_under_peak(sdata, peak=[xmin,xmax], df_index=df_index, min_as_baseline=True)
    normalized_dataframes = []
    for df_i, df in enumerate(sdata.dfs):
        area_df, _ = area_under_peak(sdata, peak=[xmin,xmax], df_index=df_i, min_as_baseline=True)
        scaling_factor = area_df0 / area_df
        df[df.columns[1]] =  df[df.columns[1]] * scaling_factor
        normalized_dataframes.append(df)
    sdata.dfs = normalized_dataframes
    return sdata


def unit_str(unit:str):
    """Normalize `unit` string from user input."""
    for key, value in alias.units.items():
        if unit in value:
            return key
    print(f"WARNING: Unknown unit '{unit}'")
    return unit




"""
# Physico-chemical constants

This subpackage contains universal physical constants,
as well as chemical data from all known elements.
It also includes functions to manage this data.


# Index

| | |
| --- | --- |
| `aton.phys.units`     | Universal constants and conversion factors |
| `aton.phys.atoms`     | Data from all chemical elements |
| `aton.phys.functions` | Functions to manage data from the `aton.phys.atoms` dict |


# Examples

All values and functions from **phys** submodules can be
loaded directly as `phys.value` or `phys.function()`,
as in the example below.

```python
from aton import phys
phys.eV_to_J                     # 1.602176634e-19
phys.atoms['H'].isotope[2].mass  # 2.0141017779
phys.split_isotope('He4')        # ('He', 4)
```

See the API reference of the specific modules for more information.


# References

## `aton.phys.units`

Constant values come from the 2022 CODATA Internationally
recommended 2022 values of the Fundamental Physical Constants.


## `aton.phys.atoms`

Atomic `mass` are in atomic mass units (amu), and come from:
Pure Appl. Chem., Vol. 78, No. 11, pp. 2051-2066, 2006.
The following masses are obtained from Wikipedia:
Ac: 227, Np: 237, Pm: 145, Tc: 98.

Isotope `mass`, `mass_number` and `abundance` come from:
J. R. de Laeter, J. K. Böhlke, P. De Bièvre, H. Hidaka, H. S. Peiser, K. J. R. Rosman
and P. D. P. Taylor (2003). *"Atomic weights of the elements. Review 2000 (IUPAC Technical Report)"*.

Total bound scattering `cross_section` $\\sigma_s$ are in barns (1 b = 100 fm$^2$).
From Felix Fernandez-Alonso, *"Neutron Scattering Fundamentals"*, 2013.

"""

from .units import *
from .functions import *
from .atoms import atoms




from aton.text import find


sample = 'tests/samples/sample.txt'


def test_lines():
    assert find.lines(filepath=sample, key='5', matches=0, additional=0, split=False, regex=False) == ['line5']
    assert find.lines(filepath=sample, key='5', matches=-1, additional=0, split=False, regex=False) == ['line5']
    assert find.lines(filepath=sample, key='5', matches=1, additional=1, split=True, regex=False) == ['line5', 'line6']
    assert find.lines(filepath=sample, key='5', matches=1, additional=-1, split=True, regex=False) == ['line4', 'line5']
    assert find.lines(filepath=sample, key='5', matches=0, additional=0, split=False, regex=True) == ['line5']
    assert find.lines(filepath=sample, key='line', matches=1, additional=0, split=False, regex=True) == ['line1']
    assert find.lines(filepath=sample, key='line', matches=-1, additional=0, split=False, regex=True) == ['line9']
    assert find.lines(filepath=sample, key='line', matches=-2, additional=1, split=True, regex=False) == ['line8', 'line9', 'line9']
    assert find.lines(filepath=sample, key='line', matches=-2, additional=1, split=True, regex=True) == ['line8', 'line9', 'line9']


def test_lines_none():
    assert find.lines(filepath=sample, key='nonexisting', matches=0, additional=0, split=False, regex=False) == []
    assert find.lines(filepath=sample, key='nonexisting', matches=0, additional=0, split=False, regex=True) == []


def test_between():
    # With and without keys
    assert find.between(filepath=sample, key1='5', key2='7', include_keys=False, match=1, regex=False) == 'line6'
    assert find.between(filepath=sample, key1='5', key2='7', include_keys=True, match=1, regex=False) == 'line5\nline6\nline7'
    assert find.between(filepath=sample, key1='5', key2='7', include_keys=False, match=1, regex=True) == 'line6'
    assert find.between(filepath=sample, key1='5', key2='7', include_keys=True, match=1, regex=True) == 'line5\nline6\nline7'
    # Nonexisting second key, it should go all the way to the end
    assert find.between(filepath=sample, key1='5', key2='nonexisting', include_keys=True, match=1, regex=False) == 'line5\nline6\nline7\nline8\nline9'
    assert find.between(filepath=sample, key1='5', key2='nonexisting', include_keys=True, match=1, regex=True) == 'line5\nline6\nline7\nline8\nline9'
    # From the end
    assert find.between(filepath=sample, key1='5', key2='nonexisting', include_keys=True, match=-1, regex=False) == 'line5\nline6\nline7\nline8\nline9'
    assert find.between(filepath=sample, key1='5', key2='nonexisting', include_keys=True, match=-1, regex=True) == 'line5\nline6\nline7\nline8\nline9'
    # With repeated keys, it should go all the way to the end
    assert find.between(filepath=sample, key1='line5', key2='line5', include_keys=True, match=1, regex=False) == 'line5\nline6\nline7\nline8\nline9'
    assert find.between(filepath=sample, key1='line5', key2='line5', include_keys=True, match=1, regex=True) == 'line5\nline6\nline7\nline8\nline9'
    assert find.between(filepath=sample, key1='line5', key2='line5', include_keys=True, match=-1, regex=False) == 'line5\nline6\nline7\nline8\nline9'
    assert find.between(filepath=sample, key1='line5', key2='line5', include_keys=True, match=-1, regex=True) == 'line5\nline6\nline7\nline8\nline9'
    # Without including keys
    assert find.between(filepath=sample, key1='5', key2='nonexisting', include_keys=False, match=-1, regex=False) == 'line6\nline7\nline8\nline9'
    assert find.between(filepath=sample, key1='5', key2='nonexisting', include_keys=False, match=-1, regex=True) == 'line6\nline7\nline8\nline9'


def test_between_none():
    assert find.between(filepath=sample, key1='nonexisting', key2='7', include_keys=False, match=1, regex=False) == ''
    assert find.between(filepath=sample, key1='nonexisting', key2='nonexisting', include_keys=False, match=1, regex=False) == ''
    assert find.between(filepath=sample, key1='nonexisting', key2='7', include_keys=False, match=1, regex=True) == ''
    assert find.between(filepath=sample, key1='nonexisting', key2='nonexisting', include_keys=False, match=1, regex=True) == ''




'''
# Description

Functions to move files around.


# Index

`get()`  
`get_list()`  
`copy()`  
`move()`  
`remove()`  
`rename_on_folder()`  
`rename_on_folders()`  
`copy_to_folders()`  

---
'''


import os
import shutil


def get(
        filepath,
        filters=None,
        return_anyway:bool=False
        ) -> str:
    '''
    Check if the given `filepath` exists in the currrent working directory
    or in the full path, and returns its full path as a string.\n
    Raises an error if the file is not found, unless `return_anyway=True`,
    in which case it returns None. This can be used to personalize errors.\n
    If the provided string is a directory, it checks the files inside it.
    if there is only one file inside, it returns said file;
    if there are more files, it tries to filter them with the `filters` keyword(s) to return a single file.
    If this fails, try using more strict filers to return a single file.
    '''
    if os.path.isfile(filepath):
        return os.path.abspath(filepath)
    elif os.path.isdir(filepath):
        files = get_list(filepath, filters, abspath=True)
    elif return_anyway:
        return None
    else:
        raise FileNotFoundError('Nothing found at ' + str(filepath))
    # Return a single file
    if len(files) == 1:
        return files[0]
    elif return_anyway:
        return None
    elif len(files) == 0:
        raise FileNotFoundError('The following directory is empty (maybe due to the filters):\n' + filepath)
    else:
        raise FileExistsError(f'More than one file found, please apply a more strict filter. Found:\n{files}')


def get_list(
        folder:str,
        filters=None,
        abspath:bool=True
    ) -> list:
    '''
    Takes a `folder`, filters the content with the `filters` keyword(s) if provided, and returns a list with the matches.
    The full paths are returned by default; to get only the base names, set `abspath=False`.
    '''
    if os.path.isfile(folder):
        folder = os.path.dirname(folder)
    if not os.path.isdir(folder):
        raise FileNotFoundError('Directory not found: ' + folder)
    folder = os.path.abspath(folder)
    files = os.listdir(folder)
    # Apply filters or not
    if filters is not None:
        target_files = []
        if not isinstance(filters, list):
            filters = [str(filters)]
        for filter_i in filters:
            filter_i = os.path.basename(filter_i)
            for f in files:
                if filter_i in f:
                    target_files.append(f)
        files = target_files
    if abspath:
        filepaths = []
        for f in files:
            filepaths.append(os.path.join(folder, f))
        files = filepaths
    return files


def copy(
        old:str,
        new:str
    ) -> None:
    """Copies the content of `old` file to `new` file with shutil"""
    file = shutil.copy(old, new)
    return None


def move(
        old:str,
        new:str
    ) -> None:
    """Moves `old` file to `new` file."""
    file = shutil.move(old, new)
    return None


def remove(filepath:str) -> None:
    '''
    Removes the given file or folder at `filepath`.
    > WARNING: Removing stuff is always dangerous, be careful!
    '''
    if filepath is None:
        return None  # It did not exist in the first place
    elif os.path.isfile(filepath):
        os.remove(filepath)
    elif os.path.isdir(filepath):
        shutil.rmtree(filepath)
    else:
        return None  # It did not exist in the first place
    return None


def rename_on_folder(
        old:str,
        new:str,
        folder=None
    ) -> None:
    '''
    Batch renames files in the given `folder`, replacing `old` string by `new` string.
    If no folder is provided, the current working directory is used.
    '''
    if folder is None:
        files = os.listdir('.')
    elif os.path.isdir(folder):
        file_list = os.listdir(folder)
        files = []
        for file in file_list:
            file_path = os.path.join(folder, file)
            files.append(file_path)
    elif os.path.isdir(os.path.join(os.getcwd(), folder)):
        folder_path = os.path.join(os.getcwd(), folder)
        file_list = os.listdir(folder_path)
        files = []
        for file in file_list:
            file_path = os.path.join(folder_path, file)
            files.append(file_path)
    else:
        raise FileNotFoundError('Missing folder at ' + folder + ' or in the CWD ' + os.getcwd())
    for f in files:
        if old in f:
            os.rename(f, f.replace(old, new))
    return None


def rename_on_folders(
        old:str,
        new:str,
        folder=None
    ) -> None:
    '''
    Renames the files inside the subfolders in the parent `folder`,
    from an `old` string to the `new` string.
    If no `folder` is provided, the current working directory is used.
    '''
    if folder is None:
        things = os.listdir('.')
    elif os.path.isdir(folder):
        things = os.listdir(folder)
    elif os.path.isdir(os.path.join(os.getcwd(), folder)):
        things = os.listdir(os.path.join(os.getcwd(), folder))
    else:
        raise FileNotFoundError('Missing folder at ' + folder + ' or in the CWD ' + os.getcwd())
    for d in things:
        if os.path.isdir(d):
            for f in os.listdir(d):
                if old in f:
                    old_file = os.path.join(d, f)
                    new_file = os.path.join(d, f.replace(old, new))
                    os.rename(old_file, new_file)
    return None


def copy_to_folders(
        extension:str=None,
        strings_to_delete:list=[],
        folder=None
    ) -> None:
    '''
    Copies the files from the parent `folder` with the given `extension` to individual subfolders.
    The subfolders are named as the original files,
    removing the strings from the `strings_to_delete` list.
    If no `folder` is provided, it runs in the current working directory.
    '''
    if folder is None:
        folder = os.getcwd()
    old_files = get_list(folder, extension)
    if old_files is None:
        raise ValueError('No ' + extension + ' files found in path!')
    for old_file in old_files:
        new_file = old_file
        for string in strings_to_delete:
            new_file = new_file.replace(string, '')
        path = new_file.replace(extension, '')
        os.makedirs(path, exist_ok=True)
        new_file_path = os.path.join(path, new_file)
        copy(old_file, new_file_path)
    return None




"""
# Description

This module contains common dictionaries to normalise and correct user inputs.
All values are in lowercase to allow comparison with the `string.lower()` method.


# Index

`units`  
`spatial`  
`experiments`  
`files`  
`boolean`  


# Examples

```python
unit = 'Electronvolts'
if unit.lower() in aton.alias.units['eV']:
    ... do stuff ...
```

---
"""


units: dict = {
    'mol'  : ['mol', 'mols', 'mole', 'moles'],
    'g'    : ['g', 'gram', 'grams'],
    'kg'   : ['kg', 'kilogram', 'kilograms'],
    'amu'  : ['amu', 'atomicmassunit', 'atomicmassunits'],
    'eV'   : ['ev', 'electronvolt', 'electronvolts'],
    'meV'  : ['mev', 'millielectronvolt', 'millielectronvolts'],
    'J'    : ['j', 'joule', 'joules'],
    'cal'  : ['cal', 'calorie', 'calories'],
    'kcal' : ['kcal', 'kilocalorie', 'kilocalories'],
    'Ry'   : ['ry', 'rydberg', 'rydbergs'],
    'cm-1' : ['cm^{-1}', 'cm1', 'cm-1', 'cm^-1'],
    'cm'   : ['cm', 'centimeter', 'centimeters'],
    'A'    : ['a', 'aa', 'angstrom', 'angstroms', 'armstrong', 'armstrongs'],
    'bohr' : ['bohr', 'bohrs', 'bohrradii'],
    'm'    : ['m', 'meter', 'meters'],
    'deg'  : ['deg', 'degree', 'degrees'],
    'rad'  : ['rad', 'radian', 'radians'],
    'bar'  : ['bar', 'bars'],
    'kbar' : ['kbar', 'kilobar', 'kilobars'],
    'Pa'   : ['pa', 'pascal', 'pascals'],
    'GPa'  : ['gpa', 'gigapascal', 'gigapascals'],
    's'    : ['s', 'second', 'seconds'],
    'H'    : ['h', 'hour', 'hours'],
}
"""Dict with unit names."""


spatial: dict = {
    'height' : ['height', 'h'],
    'area'   : ['area', 'a'],
    'volume' : ['volume', 'vol'],
    'x'      : ['x', 'horizontal', 'h'],
    'y'      : ['y', 'vertical', 'v'],
    'z'      : ['z'],
}
"""Dict with different spatial parameters."""


experiments: dict = {
    'ins'   : ['ins', 'inelasticneutronscattering', 'inelastic neutron scattering'],
    'atr'   : ['atr', 'ftir', 'attenuatedtotalreflection', 'attenuated total reflection'],
    'raman' : ['raman'],
    'qens'  : ['qens', 'quasielasticneutronscattering', 'quasielastic neutron scattering', 'quasi elastic neutron scattering'],
}
"""Dictionary with the available experiment types."""


files: dict = {
    'file'  : ['file', 'files', 'f', 'filepath', 'file path', 'filename', 'file name'],
    'dir'   : ['dir', 'directory', 'd', 'folder'],
    'error' : ['error', 'errors', 'e', 'err'],
    }
"""Strings related to files."""


boolean: dict = {
    True  : ['yes', 'YES', 'Yes', 'Y', 'y', 'T', 'True', 'TRUE', 't', 'true', True, 'Si', 'SI', 'si', 'S', 's'],
    False : ['no', 'NO', 'No', 'N', 'n', 'F', 'False', 'FALSE', 'f', 'false', False],
}
"""Strings with booleans such as 'yes' / 'no'."""




"""
# Description

This module contains premade examples of material compositions, for testing purposes.
The `aton.spx.classes.Material.grams` is yet to be provided,
before setting the material as `aton.spx.Material.set()`.

---
"""


from .classes import Material


CH3NH3PbI3 = Material(
    elements={'Pb': 1, 'I': 3, 'C': 1, 'N': 1, 'H': 6},
    name='MAPbI3'
    )
"""CH$_3$NH$_3$PbI$_3$"""


CD3ND3PbI3 = Material(
    elements={'Pb': 1, 'I': 3, 'C': 1, 'N': 1, 'H2': 6},
    name='CD3ND3PbI3',
    peaks = {
        'baseline'       : 0.057,  # IREPA-ND-02
        'baseline_error' : 0.008,
        'h6d0'           : [36.0, 39.0],
        'h5d1'           : [33.0, 35.0],
        'h4d2'           : [30.7, 33.0],
        'h3d3'           : [28.8, 30.7],
    },
)
"""CD$_3$ND$_3$PbI$_3$.

With experimental values of the partially-deuterated amine peaks
for the disrotatory mode of MAPbI3's methylammonium.
Measured at TOSCA, ISIS RAL, UK, May 2024.
"""


CH3ND3PbI3 = Material(
    elements={'Pb': 1, 'I': 3, 'C': 1, 'N': 1, 'H': 3, 'H2': 3},
    name='CH3ND3PbI3',
    peaks = {
        'baseline'       : 0.057,  # IREPA-ND-02
        'baseline_error' : 0.008,
        'h6d0'           : [36.0, 39.0],
        'h5d1'           : [33.0, 35.0],
        'h4d2'           : [30.7, 33.0],
        'h3d3'           : [28.8, 30.7],
    },
)
"""CH$_3$ND$_3$PbI$_3$.

With experimental values of the partially-deuterated amine peaks
for the disrotatory mode of MAPbI3's methylammonium.
Measured at TOSCA, ISIS RAL, UK, May 2024.
"""
#MAPI_ND.set()


CD3NH3PbI3 = Material(
    elements={'Pb': 1, 'I': 3, 'C': 1, 'N': 1, 'H': 3, 'H2': 3},
    name='CD3NH3PbI3',
)
"""CD$_3$NH$_3$PbI$_3$"""
#MAPI_CD.set()


CH3NH3I = Material(
    elements={'C' : 1, 'N': 1, 'H': 6},
    name='CH3NH3'
)
"""CH$_3$NH$_3$I"""
#CH3NH3I.set()


CH3ND3I = Material(
    elements={'C' : 1, 'N': 1, 'H': 3, 'H2': 3},
    name='CH3ND3'
)
"""CH$_3$ND$_3$I"""
#CH3ND3I.set()




"""
# Description

This module contains functions for fitting and analyzing spectral data.


# Index

| | |
| --- | --- |
| `plateau()`         | Fit the mean value and the error of a plateau |
| `area_under_peak()` | Calculate the area under a given peak |
| `ratio_areas()`     | Check the ratio between two areas |
| `mean()`            | Get the mean and standard deviation of a list of values |

---
"""


import math
import scipy
import numpy as np
from copy import deepcopy
from .classes import Spectra
import aton.st.alias as alias


def plateau(
        spectra:Spectra,
        cuts=None,
        df_index:int=0
    ) -> tuple:
    """Fit the mean value and the error of a plateau in a `aton.spectra.Spectra` object.

    Use as `aton.spx.fit.plateau(spectra, cuts=[low_cut, high_cut], df_index=0)`.

    If `aton.spx.classes.Spectra.dfs[df_index]` has an 'Error' column, those errors are also taken into account
    along with the standard deviation of the mean, else only the standard deviation is considered.
    This is the case if your dataset had a third column with the errors
    when you imported the `aton.spx.classes.Spectra` object.

    Note that `cuts`, `low_cut` and/or `top_cut` can be set to None.
    """
    df = deepcopy(spectra.dfs[df_index])
    if isinstance(cuts, list):
        low_cut = cuts[0]
        if len(cuts) > 1:
            top_cut = cuts[1]
        else:
            top_cut = None
    elif isinstance(cuts, float):  # If cuts is a float, it is considered as low_cut
        low_cut = cuts
        top_cut = None
    elif cuts is None:
        low_cut = None
        top_cut = None
    else:
        raise ValueError("plateau: cuts must be a float for the low_cut, or a list")
    if low_cut is not None:
        df = df[df[df.columns[0]] >= low_cut]
    if top_cut is not None:
        df = df[df[df.columns[0]] <= top_cut]
    mean = df[df.columns[1]].mean()
    std_mean = df[df.columns[1]].std()
    error_column = next((col for col in alias.files['error'] if col in df.columns), None)  # Get the error column title
    if error_column:
        errors = df[error_column].to_numpy()
        std_data = np.sqrt(np.sum(errors**2)) / len(errors)
        std = np.sqrt(std_data**2 + std_mean**2)
    else:
        std = std_mean
    return mean, std


def area_under_peak(
        spectra:Spectra,
        peak:list,
        df_index:int=0,
        errors_as_in_baseline:bool=True,
        min_as_baseline:bool=False
    ) -> tuple:
    """Calculate the area under a given peak.

    Peaks must be defined as `peak:list=[xmin, xmax, baseline=0, baseline_error=0]`.
    If the dataset has no `Error` column, the error for each point is assumed to be the same
    as the baseline error if `errors_as_in_baseline=True`, otherwise it is assumed to be zero.
    If `min_as_baseline=True` and `baseline=0`, the baseline is assumed to be the minimum value.
    Also, if `min_as_baseline=True` and there are negative areas even after applying the baseline,
    the baseline will be corrected to the minimum value.
    """
    if len(peak) < 2:
        raise ValueError("area_under_peak: peak must have at least two values: [xmin, xmax]")
    xmin = peak[0]
    xmax = peak[1]
    baseline = peak[2] if len(peak) >= 3 else 0.0
    baseline_error = peak[3] if len(peak) >= 4 else 0.0

    df = deepcopy(spectra.dfs[df_index])
    df_range = df[(df[df.columns[0]] >= xmin) & (df[df.columns[0]] <= xmax)]
    x = df_range[df.columns[0]].to_numpy()
    y = df_range[df.columns[1]].to_numpy()
    
    min_y = y.min()
    if min_as_baseline and (baseline == 0 or baseline > min_y):
        baseline = min_y

    y = y - baseline

    area = scipy.integrate.simpson(y, x=x)

    # Get the error column title
    error_column = next((col for col in df.columns if col.lower() in alias.files['error']), None)  
    if error_column:
        point_errors = df_range[error_column].to_numpy()
    else: # Assume the error in each point is the same as the baseline error
        if errors_as_in_baseline == True:
            point_errors = np.full_like(y, baseline_error)
        else:
            point_errors = np.zeros_like(y)
    total_errors = np.sqrt(point_errors**2 + baseline_error**2)
    area_error = np.sqrt(scipy.integrate.simpson(total_errors**2, x=x))

    return area, area_error


def ratio_areas(
        area:float,
        area_total:float,
        area_error:float=0.0,
        area_total_error:float=0.0,
        inverse_ratio:bool=False
    ) -> tuple:
    """Check the ratio between two areas, e.g. to estimate deuteration levels from ATR data.
    
    The ratio is calculated as `area / area_total`. This behavior is modified if `inverse_ratio = True`,
    so that the ratio is calculated as `(area_total - area) / area_total`.
    Note that changing the ratio calculation also affects the error propagation.
    """
    if inverse_ratio:
        ratio = (area_total - area) / area_total
        if ratio != 0.0:
            ratio_error = abs(ratio) * np.sqrt((np.sqrt(area_total_error**2 + area_error**2) / (area_total - area))**2 + (area_total_error / area_total)**2)
        else:
            ratio_error = None
    else:
        ratio = area / area_total
        if ratio != 0.0:
            ratio_error = abs(ratio) * np.sqrt((area_error / area)**2 + (area_total_error / area_total)**2)
        else:
            ratio_error = None
    
    return ratio, ratio_error


def mean(
        array:list,
        rounded:bool=True,
        degrees_of_freedom=0
    ) -> tuple:
    """Takes an `array` of numerical values and returns the mean and standard deviation.

    It is calculated with numpy as:\n
    $\\sigma_{x}=\\sqrt{\\frac{\\sum{(x_{i}-{\\overline{x}})^2}}{N-\\text{ddof}}}$\n
    where ddof are the delta `degrees_of_freedom`, zero by default.
    Set it to `1` for a corrected sample standard deviation (low N cases),
    see more details [here](https://en.wikipedia.org/wiki/Standard_deviation#Corrected_sample_standard_deviation).\n
    The mean is rounded up to the order of the error by default. To override this behaviour, set `rounded=False`.
    """
    if not all(isinstance(x, (int, float, np.ndarray)) for x in array):
        raise ValueError("mean_std(list) requires numerical values (int, float, or numpy.ndarray).")
    data = np.asarray(array)
    mean = float(data.mean())
    error = float(data.std(ddof=degrees_of_freedom))
    if not rounded or error == 0:
        return mean, error
    exponent = int(math.floor(math.log10(abs(error))))
    first_three_digits = int(100*abs(error) / 10**exponent)
    if 104 < first_three_digits < 195:
        exponent -= 1
    rounded_mean = round(mean, -exponent)
    rounded_error = round(error, -exponent)
    return rounded_mean, rounded_error




import aton


folder = 'tests/samples/'


def test_normalize_cell_params():
    cell_params = 'CELL_PARAMETERS (alat= 10.0000)\n    1.00000000000   0.000000000   0.000000000\n   0.000000000   1.000000000   0.000000000 \n 0.000000000   0.000000000   1.0000000 '
    ideal_params = [
        'CELL_PARAMETERS alat= 10.0',
        '1.000000000000000   0.000000000000000   0.000000000000000',
        '0.000000000000000   1.000000000000000   0.000000000000000',
        '0.000000000000000   0.000000000000000   1.000000000000000',]
    normalized_params = aton.interface.qe.normalize_card(cell_params)
    assert normalized_params == ideal_params
    # Now check as a list
    cell_params = cell_params.splitlines()
    # With bohr values
    cell_params[0] = r' CELL_PARAMETERS {bohr}'
    ideal_params[0] = 'CELL_PARAMETERS bohr'
    normalized_params = aton.interface.qe.normalize_card(cell_params)
    assert normalized_params == ideal_params
    # With armstrong values
    cell_params[0] = r' CELL_PARAMETERS {angstrom}'
    ideal_params[0] = 'CELL_PARAMETERS angstrom'
    normalized_params = aton.interface.qe.normalize_card(cell_params)
    assert normalized_params == ideal_params


def test_normalize_atomic_positions():
    atomic_positions = " ATOMIC_POSITIONS {crystal} \n I   5.000000   0.0000000000000   0.000000000000000 \n C   0.000000000000000   5.000000000000000000   0.000000 "
    ideal_positions = [
        'ATOMIC_POSITIONS crystal',
        'I   5.000000000000000   0.000000000000000   0.000000000000000',
        'C   0.000000000000000   5.000000000000000   0.000000000000000']
    normalized_positions = aton.interface.qe.normalize_card(atomic_positions)
    assert normalized_positions == ideal_positions


def test_normalize_atomic_species():
    atomic_species = " ATOMIC_SPECIES \n     I  126.90400   I.upf  \nHe4   4.0026032497   He.upf\n\n! C   12.01060   C.upf\n ATOMIC_POSITIONS\n '  I   5.000000000000000   0.000000000000000   0.000000000000000'"
    ideal_species = ['ATOMIC_SPECIES', 'I   126.904   I.upf', 'He4   4.0026032497   He.upf']
    normalized_species = aton.interface.qe.normalize_card(atomic_species)
    assert normalized_species == ideal_species


def test_read():
    ideal = {
        # relax.out
        'Energy'               : -1000.0,
        'Volume'               : 2.0,
        'Density'              : 1.0,
        'Alat'                 : 10,
        'BFGS converged'       : True,
        'BFGS failed'          : False,
        'Total force'          : 0.000001,
        'Total SCF correction' : 0.0,
        'ibrav'                : 1,
        'Runtime'              : '48m 8.71s',
        'CELL_PARAMETERS_out'  : [
            'CELL_PARAMETERS alat= 10.0',
            '1.000000000   0.000000000   0.000000000',
            '0.000000000   1.000000000   0.000000000',
            '0.000000000   0.000000000   1.000000000'],
        'ATOMIC_POSITIONS_out' : [
            'ATOMIC_POSITIONS crystal',
            'I                1.0000000000        0.0000000000        0.0000000000',
            'C                0.0000000000        1.0000000000        0.0000000000',
            'N                0.0000000000        0.0000000000        1.0000000000'],
        # relax.in
        'K_POINTS'             : [
            'K_POINTS automatic',
            '2 2 2 0 0 0'],
        'ecutwfc'              : 60.0,
        'etot_conv_thr'        : 1.0e-12,
        'max_seconds'          : 1000,
        'pseudo_dir'           : "'./pseudos/'",
        'CELL_PARAMETERS' : [
            'CELL_PARAMETERS alat',
            '2.000000000000000   0.000000000000000   0.000000000000000',
            '0.000000000000000   2.000000000000000   0.000000000000000',
            '0.000000000000000   0.000000000000000   2.000000000000000'],
        'ATOMIC_SPECIES'       : [
            'ATOMIC_SPECIES',
            'I  126.90400   I.upf',
            'N   14.00650   N.upf',
            'C   12.01060   C.upf'],
        'ATOMIC_POSITIONS'     : [
            'ATOMIC_POSITIONS crystal',
            'I   5.000000000000000   0.000000000000000   0.000000000000000',
            'C   0.000000000000000   5.000000000000000   0.000000000000000',
            'N   0.000000000000000   0.000000000000000   5.000000000000000'],
    }
    result = aton.interface.qe.read_dir(folder=folder, in_str='relax.in', out_str='relax.out')
    for key in ideal:
        if key in aton.interface.qe.pw_cards:
            ideal[key] = aton.interface.qe.normalize_card(ideal[key])
        assert result[key] == ideal[key]


def test_scf_from_relax():
    ideal = {
        'calculation'      : "'scf'",
        'celldm(1)'        : 10.0,
        'ibrav'            : 0,
        'occupations'      : "'fixed'",
        'conv_thr'         : 2.0e-12,
        'ATOMIC_SPECIES'   : [
            'ATOMIC_SPECIES',
            'I  126.90400   I.upf',
            'N   14.00650   N.upf',
            'C   12.01060   C.upf'],
        'CELL_PARAMETERS'  : [
            'CELL_PARAMETERS alat',
            '1.000000000   0.000000000   0.000000000',
            '0.000000000   1.000000000   0.000000000',
            '0.000000000   0.000000000   1.000000000'],
        'ATOMIC_POSITIONS' : [
            'ATOMIC_POSITIONS crystal',
            'I                1.0000000000        0.0000000000        0.0000000000',
            'C                0.0000000000        1.0000000000        0.0000000000',
            'N                0.0000000000        0.0000000000        1.0000000000'],
    }
    aton.interface.qe.scf_from_relax(folder=folder)
    result = aton.interface.qe.read_in(folder + 'scf.in')
    for key in ideal:
        if key in ['ATOMIC_SPECIES', 'CELL_PARAMETERS', 'CELL_PARAMETERS_out', 'ATOMIC_POSITIONS', 'ATOMIC_POSITIONS_out']:
            ideal[key] = aton.interface.qe.normalize_card(ideal[key])
        assert result[key] == ideal[key]
    assert 'A' not in result.keys()
    try:
        aton.st.file.remove(folder + 'scf.in')
    except:
        pass


def test_update_other_values():
    tempfile = folder + 'temp.in'
    aton.st.file.copy(folder + 'relax.in', tempfile)
    aton.interface.qe.set_value(tempfile, 'celldm(1)', 10.0)
    modified = aton.interface.qe.read_in(tempfile)
    assert 'A' not in modified.keys()
    aton.st.file.remove(tempfile)


def test_set_value():
    tempfile = folder + 'temp.in'
    aton.st.file.copy(folder + 'relax.in', tempfile)
    aton.interface.qe.set_value(tempfile, 'ecutwfc', 80.0)
    aton.interface.qe.set_value(tempfile, 'ibrav', 5)
    aton.interface.qe.set_value(tempfile, 'calculation', "'vc-relax'")
    aton.interface.qe.set_value(tempfile, 'celldm(1)', 10.0)
    modified = aton.interface.qe.read_in(tempfile)
    # Check some unmodified values
    assert modified['max_seconds'] == 1000
    assert modified['input_dft'] == "'PBEsol'"
    # Check the modified
    assert 'A' not in modified.keys()
    assert modified['ecutwfc'] == 80.0
    assert modified['ibrav'] == 5
    assert modified['calculation'] == "'vc-relax'"
    assert modified['celldm(1)'] == 10.0
    aton.interface.qe.set_value(tempfile, 'celldm(1)', '')
    modified = aton.interface.qe.read_in(tempfile)
    assert 'A' not in modified.keys()
    assert 'celldm(1)' not in modified.keys()
    aton.st.file.remove(tempfile)


def test_count_elements():
    atomic_positions = [
        'ATOMIC_POSITIONS crystal',
        'I   5.0000000000        0.0000000000        0.0000000000',
        'C   0.0000000000        5.0000000000        0.0000000000',
        'N   0.0000000000        0.0000000000        5.0000000000',
        'Cl   0.0  0.0  0.0',
        'Cl  1.0  1.0  1.0']
    ideal = {'I': 1, 'C': 1, 'N': 1, 'Cl': 2}
    obtained = aton.interface.qe.count_elements(atomic_positions)
    for key in ideal.keys():
        assert ideal[key] == obtained[key]
    # Again, in case it does something weird
    obtained = aton.interface.qe.count_elements(atomic_positions)
    for key in ideal.keys():
        assert ideal[key] == obtained[key]


def test_add_atom():
    ideal_positions = [
        'ATOMIC_POSITIONS crystal',
        'I                5.0000000000        0.0000000000        0.0000000000',
        'C                0.0000000000        5.0000000000        0.0000000000',
        'N                0.0000000000        0.0000000000        5.0000000000',
        'O   0.0  0.0  0.0',
        'Cl  1.0  1.0  1.0']
    ideal_positions = aton.interface.qe.normalize_card(ideal_positions)
    tempfile = folder + 'temp.in'
    aton.st.file.copy(folder + 'relax.in', tempfile)
    position_1 = '  O   0.0   0.0   0.0'
    position_2 = ['Cl', 1.0, 1.0, 1.0]
    aton.interface.qe.add_atom(filepath=tempfile, position=position_1)
    aton.interface.qe.add_atom(filepath=tempfile, position=position_2)
    temp = aton.interface.qe.read_in(tempfile)
    nat = temp['nat']
    ntyp = temp['ntyp']
    atomic_positions = temp['ATOMIC_POSITIONS']
    assert nat == 5
    assert ntyp == 5
    number_of_elements = aton.interface.qe.count_elements(atomic_positions)
    ideal_dict = {'I':1, 'C':1, 'N':1, 'O':1, 'Cl':1}
    for key in ideal_dict.keys():
        assert ideal_dict[key] == number_of_elements[key]
    # Assert we have the same ATOMIC_POSITIONS
    for i, ideal in enumerate(ideal_positions):
        ideal_str = ideal.split()
        detected_str = atomic_positions[i].split()
        assert detected_str == ideal_str
    # Additional surrounding values, just in case
    assert temp['ibrav'] == 1
    assert temp['A'] == 10.0
    assert temp['ecutwfc'] == 60.0
    assert temp['input_dft'] == "'PBEsol'"
    aton.st.file.remove(tempfile)


def test_get_atom():
    relax = folder + 'relax.in'
    ideal = 'N   0.000000000000000   0.000000000000000   5.000000000000000'
    approx_list_1 = [0.00, 0.00, 5.01]
    approx_list_2 = [0.0, 0.0, 5.1]
    approx_list_3 = [0.0, 0.00, 5.09]
    approx_str = '0.00, 0.00, 5.01'
    assert aton.interface.qe.get_atom(relax, approx_list_1) == ideal
    assert aton.interface.qe.get_atom(relax, approx_list_2) == ideal
    assert aton.interface.qe.get_atom(relax, approx_list_3) == ideal
    assert aton.interface.qe.get_atom(relax, approx_str) == ideal

